"""Functions to find and extract data from a catalog."""

import datetime
import logging
import os
import re
import warnings
from collections import defaultdict
from collections.abc import Callable, Sequence
from copy import deepcopy
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr
import xclim as xc
from intake_esm.derived import DerivedVariableRegistry
from xclim.core.calendar import compare_offsets

from .catalog import (
    ID_COLUMNS,
    DataCatalog,
    concat_data_catalogs,
    generate_id,
    subset_file_coverage,
)
from .catutils import parse_from_ds
from .config import parse_config
from .indicators import load_xclim_module, registry_from_module
from .spatial import subset
from .utils import CV, _xarray_defaults
from .utils import ensure_correct_time as _ensure_correct_time
from .utils import get_cat_attrs, natural_sort, standardize_periods, xrfreq_to_timedelta

logger = logging.getLogger(__name__)


__all__ = [
    "extract_dataset",
    "get_period_from_warming_level",
    "get_warming_level",
    "get_warming_level_from_period",
    "resample",
    "search_data_catalogs",
    "subset_warming_level",
]


@parse_config
def extract_dataset(  # noqa: C901
    catalog: DataCatalog,
    *,
    variables_and_freqs: dict | None = None,
    periods: list[str | int] | list[list[str | int]] | None = None,
    region: dict | None = None,
    to_level: str = "extracted",
    ensure_correct_time: bool = True,
    xr_open_kwargs: dict | None = None,
    xr_combine_kwargs: dict | None = None,
    preprocess: Callable | None = None,
    resample_methods: dict | None = None,
    mask: bool | xr.Dataset | xr.DataArray = False,
) -> dict:
    """
    Take one element of the output of `search_data_catalogs` and returns a dataset,
    performing conversions and resampling as needed.

    Nothing is written to disk within this function.

    Parameters
    ----------
    catalog : DataCatalog
        Sub-catalog for a single dataset, one value of the output of `search_data_catalogs`.
    variables_and_freqs : dict, optional
        Variables and frequencies, following a 'variable: xrfreq-compatible str' format. A list of strings can also be provided.
        If None, it will be read from `catalog._requested_variables` and `catalog._requested_variable_freqs`
        (set by `variables_and_freqs` in `search_data_catalogs`).
    periods : list of str or list of int or list of lists of str or list of lists of int, optional
        Either [start, end] or list of [start, end] for the periods to be evaluated.
        Will be read from catalog._requested_periods if None. Leave both None to extract everything.
    region : dict, optional
        Description of the region and the subsetting method (required fields listed in the Notes) used in `xscen.spatial.subset`.
    to_level : str
        The processing level to assign to the output. Defaults to 'extracted'.
    ensure_correct_time : bool
        When True (default), even if the data has the correct frequency, its time coordinate is
        checked so that it exactly matches the frequency code (xrfreq).
        For example, daily data given at noon would be transformed to be given at midnight.
        If the time coordinate is invalid, it raises an error.
    xr_open_kwargs : dict, optional
        A dictionary of keyword arguments to pass to `DataCatalogs.to_dataset_dict`, which
        will be passed to `xr.open_dataset`.
    xr_combine_kwargs : dict, optional
        A dictionary of keyword arguments to pass to `DataCatalogs.to_dataset_dict`, which
        will be passed to `xr.combine_by_coords`.
    preprocess : callable, optional
        If provided, call this function on each dataset prior to aggregation.
    resample_methods : dict, optional
        Dictionary where the keys are the variables and the values are the resampling method.
        Options for the resampling method are {'mean', 'min', 'max', 'sum', 'wind_direction'}.
        If the method is not given for a variable, it is guessed from the variable name and frequency,
        using the mapping in CVs/resampling_methods.json.
        If the variable is not found there, "mean" is used by default.
    mask : xr.Dataset or xr.DataArray or bool
        A mask that is applied to all variables and only keeps data where it is True.
        Where the mask is False, variable values are replaced by NaNs.
        The mask should have the same dimensions as the variables extracted.
        If `mask` is a dataset, the dataset should have a variable named 'mask'.
        If `mask` is True, it will expect a `mask` variable at xrfreq `fx` to have been extracted.

    Returns
    -------
    dict
        Dictionary (keys = xrfreq) with datasets containing all available and computed variables,
        subsetted to the region, everything resampled to the requested frequency.

    Notes
    -----
    'region' fields:
        name: str
            Region name used to overwrite domain in the catalog.
        method: str
            ['gridpoint', 'bbox', shape', 'sel']
        tile_buffer: float, optional
            Multiplier to apply to the model resolution.
        kwargs
            Arguments specific to the method used.

    See Also
    --------
    intake_esm.core.esm_datastore.to_dataset_dict, xarray.open_dataset, xarray.combine_by_coords
    """
    resample_methods = resample_methods or {}

    # Checks
    # must have all the same processing level and same id
    unique = catalog.unique()
    if len(unique["processing_level"]) > 1 or len(unique["id"]) > 1:
        raise ValueError(
            "The extraction catalog must have a unique id and unique processing_level."
        )
    if region is None and len(unique["domain"]) > 1:
        raise ValueError(
            "If a subset region is not given, the extraction catalog must have a unique domain."
        )

    if variables_and_freqs is None:
        try:
            variables_and_freqs = defaultdict(list)
            for a, b in zip(
                catalog._requested_variables_true, catalog._requested_variable_freqs
            ):
                variables_and_freqs[a].extend([b])
        except ValueError:
            raise ValueError("Failed to determine the requested variables and freqs.")
    else:
        # Make everything a list
        variables_and_freqs = {
            k: [v] if not isinstance(v, list) else v
            for k, v in variables_and_freqs.items()
        }

    # Default arguments to send xarray
    xr_kwargs = _xarray_defaults(
        xr_open_kwargs=xr_open_kwargs or {}, xr_combine_kwargs=xr_combine_kwargs or {}
    )

    # Open the catalog
    ds_dict = catalog.to_dataset_dict(
        preprocess=preprocess,
        # Only print a progress bar when it is minimally useful
        progressbar=(len(catalog.keys()) > 1),
        **xr_kwargs,
    )

    out_dict = {}
    for xrfreq in np.unique([x for y in variables_and_freqs.values() for x in y]):
        ds = xr.Dataset()
        attrs = {}
        # iterate on the datasets, in reverse timedelta order
        for key, ds_ts in sorted(
            ds_dict.items(),
            key=lambda kv: pd.Timedelta(
                CV.xrfreq_to_timedelta(catalog[kv[0]].df.xrfreq.iloc[0], default="NAN")
            ),
            reverse=True,
        ):
            if "time" in ds_ts:
                if pd.Timedelta(
                    CV.xrfreq_to_timedelta(catalog[key].df.xrfreq.iloc[0])
                ) > pd.Timedelta(CV.xrfreq_to_timedelta(xrfreq)):
                    continue
                if ensure_correct_time:
                    # Expected freq (xrfreq is the wanted freq)
                    expfreq = catalog[key].df.xrfreq.iloc[0]
                    ds_ts = _ensure_correct_time(ds_ts, expfreq)

            for var_name, da in ds_ts.data_vars.items():
                # Support for grid_mapping, crs, and other such variables
                if len(da.dims) == 0 and var_name not in ds:
                    if bool(re.search("^|S[0-9]{1,4}$", str(da.dtype))):
                        da = da.astype("U")
                    ds = ds.assign({var_name: da})
                    continue

                # TODO: 2nd part is a temporary fix until this is changed in intake_esm
                if (
                    var_name in ds
                    or xrfreq not in variables_and_freqs.get(var_name, [])
                    or var_name not in catalog._requested_variables_true
                ):
                    continue

                # TODO: This is a temporary fix for an xclim bug where the grid_mapping attribute is not transferred upon calculation
                if (
                    var_name in catalog.derivedcat
                    and len(
                        set(ds_ts.data_vars).intersection(
                            catalog.derivedcat[var_name].query["variable"]
                        )
                    )
                    >= 1
                ):
                    grid_mapping = np.unique(
                        [
                            ds_ts[v].attrs["grid_mapping"]
                            for v in set(ds_ts.data_vars).intersection(
                                catalog.derivedcat[var_name].query["variable"]
                            )
                            if "grid_mapping" in ds_ts[v].attrs
                        ]
                    )
                    if len(grid_mapping) == 1:
                        da.attrs["grid_mapping"] = grid_mapping[0]
                    elif len(grid_mapping) > 1:
                        raise ValueError("Multiple grid_mapping detected.")

                if "time" not in da.dims or (
                    catalog[key].df["xrfreq"].iloc[0] == xrfreq
                ):
                    ds = ds.assign({var_name: da})
                else:  # check if it needs resampling
                    if pd.to_timedelta(
                        CV.xrfreq_to_timedelta(catalog[key].df["xrfreq"].iloc[0])
                    ) < pd.to_timedelta(CV.xrfreq_to_timedelta(xrfreq)):
                        msg = f"Resampling {var_name} from [{catalog[key].df['xrfreq'].iloc[0]}] to [{xrfreq}]."
                        logger.info(msg)
                        ds = ds.assign(
                            {
                                var_name: resample(
                                    da,
                                    xrfreq,
                                    ds=ds_ts,
                                    method=resample_methods.get(var_name, None),
                                )
                            }
                        )
                    else:
                        raise ValueError(
                            "Variable is at a coarser frequency than requested."
                        )

                attrs = ds_ts.attrs

        ds.attrs = attrs
        if "time" not in ds.dims:
            ds.attrs["cat:frequency"] = "fx"
            ds.attrs["cat:xrfreq"] = "fx"
        else:
            ds.attrs["cat:xrfreq"] = xrfreq
            ds.attrs["cat:frequency"] = CV.xrfreq_to_frequency(xrfreq)

        # Subset time on the periods
        periods_extract = deepcopy(periods)
        if periods_extract is None and hasattr(catalog, "_requested_periods"):
            periods_extract = catalog._requested_periods
        if periods_extract is not None and "time" in ds:
            periods_extract = standardize_periods(periods_extract)
            slices = []
            for period in periods_extract:
                slices.extend([ds.sel({"time": slice(period[0], period[1])})])
            ds = xr.concat(
                slices, dim="time", **xr_kwargs["xarray_combine_by_coords_kwargs"]
            )

        # subset to the region
        if region is not None:
            ds = subset(ds, **region)

        # add relevant attrs
        ds.attrs["cat:processing_level"] = to_level
        if "cat:variable" not in ds.attrs:
            ds.attrs["cat:variable"] = parse_from_ds(ds, ["variable"])["variable"]

        out_dict[xrfreq] = ds

    if mask:
        if isinstance(mask, xr.Dataset):
            ds_mask = mask["mask"]
        elif isinstance(mask, xr.DataArray):
            ds_mask = mask
        elif (
            "fx" in out_dict and "mask" in out_dict["fx"]
        ):  # get mask that was extracted above
            ds_mask = out_dict["fx"]["mask"].copy()
        else:
            raise ValueError(
                "No mask found. Either pass a xr.Dataset/xr.DataArray to the `mask` argument "
                "or pass a `dc` that includes a dataset with a variable named `mask`."
            )

        # iter over all xrfreq to apply the mask
        for xrfreq, ds in out_dict.items():
            out_dict[xrfreq] = ds.where(ds_mask)
            if xrfreq == "fx":  # put back the mask
                out_dict[xrfreq]["mask"] = ds_mask

    return out_dict


def resample(  # noqa: C901
    da: xr.DataArray,
    target_frequency: str,
    *,
    ds: xr.Dataset | None = None,
    method: str | None = None,
    missing: str | dict | None = None,
) -> xr.DataArray:
    """Aggregate variable to the target frequency.

    If the input frequency is greater than a week, the resampling operation is weighted by
    the number of days in each sampling period.

    Parameters
    ----------
    da : xr.DataArray
        DataArray of the variable to resample, must have a "time" dimension and be of a
        finer temporal resolution than "target_frequency".
    target_frequency : str
        The target frequency/freq str, must be one of the frequency supported by xarray.
    ds : xr.Dataset, optional
        The "wind_direction" resampling method needs extra variables, which can be given here.
    method : {'mean', 'min', 'max', 'sum', 'wind_direction'}, optional
        The resampling method. If None (default), it is guessed from the variable name and frequency,
        using the mapping in CVs/resampling_methods.json. If the variable is not found there,
        "mean" is used by default.
    missing: {'mask', 'drop'} or dict, optional
        If 'mask' or 'drop', target periods that would have been computed from fewer timesteps than expected
        are masked or dropped, using a threshold of 5% of missing data.
        E.g. the first season of a `target_frequency` of "QS-DEC" will be masked or dropped if data starts in January.
        If a dict, points to a xclim check missing method which will mask periods according to the number of NaN values.
        The dict must contain a "method" field corresponding to the xclim method name and may contain
        any other args to pass. Options are documented in :py:mod:`xclim.core.missing`.

    Returns
    -------
    xr.DataArray
        Resampled variable
    """
    var_name = da.name

    initial_frequency = xr.infer_freq(da.time.dt.round("min")) or "undetected"
    if initial_frequency == "undetected":
        warnings.warn(
            "Could not infer the frequency of the dataset. Be aware that this might result in erroneous manipulations."
        )

    if method is None:
        if (
            target_frequency in CV.resampling_methods.dict
            and var_name in CV.resampling_methods.dict[target_frequency]
        ):
            method = CV.resampling_methods(target_frequency)[var_name]
            msg = f"Resampling method for {var_name}: '{method}', based on variable name and frequency."
            logger.info(msg)

        elif var_name in CV.resampling_methods.dict["any"]:
            method = CV.resampling_methods("any")[var_name]
            msg = (
                f"Resampling method for {var_name}: '{method}', based on variable name."
            )
            logger.info(msg)

        else:
            method = "mean"
            msg = f"Resampling method for {var_name} defaulted to: 'mean'."
            logger.info(msg)

    weights = None
    if (
        initial_frequency != "undetected"
        and compare_offsets(initial_frequency, ">", "W")
        and method in ["mean", "median", "std", "var", "wind_direction"]
    ):
        # More than a week -> non-uniform sampling length!
        t = xr.date_range(
            da.indexes["time"][0],
            periods=da.time.size + 1,
            freq=initial_frequency,
            calendar=da.time.dt.calendar,
        )
        # This is the number of days in each sampling period
        days_per_step = (
            xr.DataArray(t, dims=("time",), coords={"time": t})
            .diff("time", label="lower")
            .dt.days
        )
        days_per_period = (
            days_per_step.resample(time=target_frequency)
            .sum()  # Total number of days per period
            .sel(time=days_per_step.time, method="ffill")  # Up-sample to initial freq
            .assign_coords(
                time=days_per_step.time
            )  # Not sure why we need this, but time coord is from the resample even after sel
        )
        weights = days_per_step / days_per_period

    # TODO : Support non-surface wind?
    if method == "wind_direction":
        ds[var_name] = da
        if ds is None or not any(
            [
                all(v in ds for v in ["uas", "vas"]),
                all(v in ds for v in ["sfcWind", "sfcWindfromdir"]),
            ]
        ):
            raise ValueError(
                "Resampling method 'wind_direction' failed to find all required variables."
            )

        # The method requires uas, vas, and sfcWind. Acquire all of them.
        if all(v in ds for v in ["uas", "vas"]):
            uas, vas = ds.uas, ds.vas
        else:
            uas, vas = xc.indicators.atmos.wind_vector_from_speed(
                ds.sfcWind, ds.sfcWindfromdir
            )
        if "sfcWind" not in ds:
            ds["sfcWind"], _ = xc.indicators.atmos.wind_speed_from_vector(
                uas=ds["uas"], vas=ds["vas"]
            )

        # Resample first to find the average wind speed and components
        if weights is not None:
            with xr.set_options(keep_attrs=True):
                ds = (ds * weights).resample(time=target_frequency).sum(dim="time")
        else:
            ds = ds.resample(time=target_frequency).mean(dim="time", keep_attrs=True)

        # Based on Vector Magnitude and Direction equations
        # For example: https://www.khanacademy.org/math/precalculus/x9e81a4f98389efdf:vectors/x9e81a4f98389efdf:component-form/a/vector-magnitude-and-direction-review  # noqa: E501

        uas_attrs = ds["uas"].attrs
        vas_attrs = ds["vas"].attrs

        # Using the direction angle theta and the vector magnitude (sfcWind), re-adjust uas/vas
        theta = np.arctan2(ds["vas"], ds["uas"])
        ds["uas"] = ds["sfcWind"] * np.cos(theta)
        ds["uas"].attrs = uas_attrs
        ds["vas"] = ds["sfcWind"] * np.sin(theta)
        ds["vas"].attrs = vas_attrs

        # Prepare output
        if var_name in ["sfcWindfromdir"]:
            _, out = xc.indicators.atmos.wind_speed_from_vector(uas=uas, vas=vas)
        else:
            out = ds[var_name]

    elif weights is not None:
        if method == "mean":
            # Avoiding resample().map() is much more performant
            with xr.set_options(keep_attrs=True):
                out = (
                    (da * weights)
                    .resample(time=target_frequency)
                    .sum(dim="time")
                    .rename(da.name)
                )
        else:
            kws = {"q": 0.5} if method == "median" else {}
            ds = xr.merge([da, weights.rename("weights")])
            out = ds.resample(time=target_frequency).map(
                lambda grp: getattr(
                    grp.drop_vars("weights").weighted(grp.weights),
                    method if method != "median" else "quantile",
                )(dim="time", **kws)
            )[da.name]
    else:
        out = getattr(da.resample(time=target_frequency), method)(
            dim="time", keep_attrs=True
        )

    missing_note = " "
    initial_td = xrfreq_to_timedelta(initial_frequency)
    if missing in ["mask", "drop"] and not pd.isnull(initial_td):
        steps_per_period = (
            xr.ones_like(da.time, dtype="int").resample(time=target_frequency).sum()
        )
        t = xr.date_range(
            steps_per_period.indexes["time"][0],
            periods=steps_per_period.time.size + 1,
            freq=target_frequency,
        )

        expected = (
            xr.DataArray(t, dims=("time",), coords={"time": t}).diff(
                "time", label="lower"
            )
            / initial_td
        )
        complete = (steps_per_period / expected) > 0.95
        action = "masking" if missing == "mask" else "dropping"
        missing_note = f", {action} incomplete periods "
    elif isinstance(missing, dict):
        missmeth = missing.pop("method")
        miss = xc.core.missing.MISSING_METHODS[missmeth](**missing)
        complete = ~miss(da, target_frequency, initial_frequency)
        missing = "mask"
        missing_note = f", masking incomplete periods according to {miss} "
    if missing in {"mask", "drop"}:
        out = out.where(complete, drop=(missing == "drop"))

    new_history = (
        f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] "
        f"{'weighted' if weights is not None else ''} {method} "
        f"resample from {initial_frequency} to {target_frequency}"
        f"{missing_note}- xarray v{xr.__version__}"
    )
    history = (
        new_history + " \n " + out.attrs["history"]
        if "history" in out.attrs
        else new_history
    )
    out.attrs["history"] = history

    return out


@parse_config
def search_data_catalogs(  # noqa: C901
    data_catalogs: (
        str | os.PathLike | DataCatalog | list[str | os.PathLike | DataCatalog]
    ),
    variables_and_freqs: dict,
    *,
    other_search_criteria: dict | None = None,
    exclusions: dict | None = None,
    match_hist_and_fut: bool = False,
    periods: list[str] | list[list[str]] | None = None,
    coverage_kwargs: dict | None = None,
    id_columns: list[str] | None = None,
    allow_resampling: bool = False,
    allow_conversion: bool = False,
    conversion_yaml: str | None = None,
    restrict_resolution: str | None = None,
    restrict_members: dict | None = None,
    restrict_warming_level: dict | bool | None = None,
) -> dict:
    """Search through DataCatalogs.

    Parameters
    ----------
    data_catalogs : str, os.PathLike, DataCatalog, or a list of those
        DataCatalog (or multiple, in a list) or paths to JSON/CSV data catalogs.
        They must use the same columns and aggregation options.
    variables_and_freqs : dict
        Variables and freqs to search for, following a 'variable: xr-freq-compatible-str' format.
        A list of strings can also be provided.
    other_search_criteria : dict, optional
        Other criteria to search for in the catalogs' columns, following a 'column_name: list(subset)' format.
        You can also pass 'require_all_on: list(columns_name)' in order to only return results that correspond to
        all other criteria across the listed columns.
        More details available at https://intake-esm.readthedocs.io/en/stable/how-to/enforce-search-query-criteria-via-require-all-on.html .
    exclusions : dict, optional
        Same as other_search_criteria, but for eliminating results.
        Any result that matches any of the exclusions will be removed.
    match_hist_and_fut: bool
        If True, historical and future simulations will be combined into the same line,
        and search results lacking one of them will be rejected.
    periods : list of str or list of lists of str, optional
        Either [start, end] or list of [start, end] for the periods to be evaluated.
    coverage_kwargs : dict, optional
        Arguments to pass to subset_file_coverage (only used when periods is not None).
    id_columns : list, optional
        List of columns used to create a id column. If None is given, the original
        "id" is left.
    allow_resampling : bool
         If True (default), variables with a higher time resolution than requested are considered.
    allow_conversion : bool
        If True (default) and if the requested variable cannot be found, intermediate variables are
        searched given that there exists a converting function in the "derived variable registry".
    conversion_yaml : str, optional
        Path to a YAML file that defines the possible conversions (used alongside 'allow_conversion'=True).
        This file should follow the xclim conventions for building a virtual module.
        If None, the "derived variable registry" will be defined by the file in "xscen/xclim_modules/conversions.yml"
    restrict_resolution : str, optional
        Used to restrict the results to the finest/coarsest resolution available for a given simulation.
        ['finest', 'coarsest'].
    restrict_members : dict, optional
        Used to restrict the results to a given number of members for a given simulation.
        Currently only supports {"ordered": int} format.
    restrict_warming_level : bool or dict, optional
        Used to restrict the results only to datasets that exist in the csv used to compute warming levels in `subset_warming_level`.
        If True, this will only keep the datasets that have a mip_era, source, experiment and member combination that exist in the csv.
        This does not guarantee that a given warming level will be reached, only that the datasets have corresponding columns in the csv.
        More option can be added by passing a dictionary instead of a boolean.
        If {'ignore_member':True}, it will disregard the member when trying to match the dataset to a column.
        If {tas_src: Path_to_netcdf}, it will use an alternative netcdf instead of the default one provided by xscen.
        If 'wl' is a provided key, then `xs.get_period_from_warming_level` will be called
        and only datasets that reach the warming level will be kept.
        This can be combined with other arguments of the function, for example {'wl': 1.5, 'window': 30}.

    Notes
    -----
    - The "other_search_criteria" and "exclusions" arguments accept wildcard (*) and regular expressions.
    - Frequency can be wildcarded with 'NA' in the `variables_and_freqs` dict.
    - Variable names cannot be wildcarded, they must be CMIP6-standard.

    Returns
    -------
    dict
        Keys are the id and values are the DataCatalogs for each entry.
        A single DataCatalog can be retrieved with `concat_data_catalogs(*out.values())`.
        Each DataCatalog has a subset of the derived variable registry that corresponds
        to the needs of this specific group.
        Usually, each entry can be written to file in a single Dataset when using
        `extract_dataset` with the same arguments.

    See Also
    --------
    intake_esm.core.esm_datastore.search
    """
    # Cast single items to a list
    if isinstance(data_catalogs, str | os.PathLike | DataCatalog):
        data_catalogs = [data_catalogs]

    # Open the catalogs given as paths
    data_catalogs = [
        (
            dc
            if not isinstance(dc, str | os.PathLike)
            else (
                DataCatalog(dc)
                if Path(dc).suffix == ".json"
                else DataCatalog.from_df(dc)
            )
        )
        for dc in data_catalogs
    ]

    if not all(isinstance(dc, DataCatalog) for dc in data_catalogs):
        raise ValueError("Catalogs type not recognized.")

    cat_kwargs = {}
    if allow_conversion:
        if conversion_yaml is None:
            conversion_yaml = Path(__file__).parent / "xclim_modules" / "conversions"
        cat_kwargs = {
            "registry": registry_from_module(load_xclim_module(conversion_yaml))
        }

    # Prepare a unique catalog to search from, with the DerivedCat added if required
    catalog = DataCatalog(
        {
            "esmcat": data_catalogs[0].esmcat.model_dump(),
            "df": pd.concat([dc.df for dc in data_catalogs], ignore_index=True),
        },
        **cat_kwargs,
    )
    msg = f"Catalog opened: {catalog} from {len(data_catalogs)} files."
    logger.info(msg)

    if match_hist_and_fut:
        logger.info("Dispatching historical dataset to future experiments.")
        catalog = _dispatch_historical_to_future(catalog, id_columns)

    # Cut entries that do not match search criteria
    if exclusions:
        for k in exclusions.keys():
            ex = catalog.search(**{k: exclusions[k]})
            catalog.esmcat._df = pd.concat([catalog.df, ex.df]).drop_duplicates(
                keep=False
            )
            msg = f"Removing {len(ex.df)} assets based on exclusion dict '{k}': {exclusions[k]}."
            logger.info(msg)
    full_catalog = deepcopy(catalog)  # Used for searching for fixed fields
    if other_search_criteria:
        catalog = catalog.search(**other_search_criteria)
        msg = (
            f"{len(catalog.df)} assets matched the criteria : {other_search_criteria}."
        )
        logger.info(msg)
    if restrict_warming_level:
        if isinstance(restrict_warming_level, bool):
            restrict_warming_level = {}
        restrict_warming_level.setdefault("ignore_member", False)
        catalog.esmcat._df = _restrict_wl(catalog.df, restrict_warming_level)

    if id_columns is not None or catalog.df["id"].isnull().any():
        ids = generate_id(catalog.df, id_columns)
        if id_columns is not None:
            # Recreate id from user specifications
            catalog.df["id"] = ids
        else:
            # Only fill in the missing IDs.
            # Unreachable line if 'id' is in the aggregation control columns, but this is a safety measure.
            catalog.df["id"] = catalog.df["id"].fillna(ids)

    if catalog.df.empty:
        warnings.warn(
            "Found no match corresponding to the search criteria.",
            UserWarning,
            stacklevel=1,
        )
        return {}

    coverage_kwargs = coverage_kwargs or {}
    periods = standardize_periods(periods, out_dtype="datetime")

    msg = f"Iterating over {len(catalog.unique('id'))} potential datasets."
    logger.info(msg)
    # Loop on each dataset to assess whether they have all required variables
    # And select best freq/timedelta for each
    catalogs = {}
    if len(catalog) > 0:
        for (sim_id,), scat in catalog.iter_unique("id"):
            # Find all the entries that match search parameters
            varcats = []
            for var_id, xrfreqs in variables_and_freqs.items():
                if isinstance(xrfreqs, str):
                    xrfreqs = [xrfreqs]
                for xrfreq in xrfreqs:
                    if xrfreq == "fx":
                        varcat = scat.search(
                            xrfreq=xrfreq,
                            variable=var_id,
                            require_all_on=["id", "xrfreq"],
                        )
                        if len(varcat) == 0:
                            # Try searching in other experiments or members
                            scat_id = {
                                i: scat.df[i].iloc[0]
                                for i in id_columns or ID_COLUMNS
                                if (
                                    (i in scat.df.columns)
                                    and (not pd.isnull(scat.df[i].iloc[0]))
                                )
                            }
                            scat_id.pop("experiment", None)
                            scat_id.pop("member", None)
                            varcat = full_catalog.search(
                                **scat_id,
                                xrfreq=xrfreq,
                                variable=var_id,
                                require_all_on=["id", "xrfreq"],
                            )
                            if len(varcat) > 1:
                                varcat.esmcat._df = varcat.df.iloc[[0]]
                            if len(varcat) == 1:
                                warnings.warn(
                                    f"Dataset {sim_id} doesn't have the fixed field {var_id}, but it can be acquired from {varcat.df['id'].iloc[0]}.",
                                    UserWarning,
                                    stacklevel=1,
                                )
                                for i in {"member", "experiment", "id"}.intersection(
                                    varcat.df.columns
                                ):
                                    varcat.df.loc[:, i] = scat.df[i].iloc[0]

                        # TODO: Temporary fix until this is changed in intake_esm
                        varcat._requested_variables_true = [var_id]
                        varcat._dependent_variables = list(
                            set(varcat._requested_variables).difference(
                                varcat._requested_variables_true
                            )
                        )
                    else:
                        # TODO: Add support for DerivedVariables that themselves require DerivedVariables
                        # TODO: Add support for DerivedVariables that exist on different frequencies (e.g. 1hr 'pr' & 3hr 'tas')
                        varcat = scat.search(
                            variable=var_id, require_all_on=["id", "xrfreq"]
                        )
                        msg = f"At var {var_id}, after search cat has {varcat.derivedcat.keys()}"
                        logger.debug(msg)
                        # TODO: Temporary fix until this is changed in intake_esm
                        varcat._requested_variables_true = [var_id]
                        varcat._dependent_variables = list(
                            set(varcat._requested_variables).difference(
                                varcat._requested_variables_true
                            )
                        )

                        # We want to match lines with the correct freq,
                        # IF allow_resampling is True and xrfreq translates to a timedelta,
                        # we also want those with (stricyly) higher temporal resolution
                        same_frq = varcat.df.xrfreq == xrfreq
                        td = pd.to_timedelta(CV.xrfreq_to_timedelta(xrfreq))
                        varcat.df["timedelta"] = pd.to_timedelta(
                            varcat.df.xrfreq.apply(
                                CV.xrfreq_to_timedelta, default="NAN"
                            )
                        )
                        # else is joker (any timedelta)
                        lower_frq = (
                            np.less(varcat.df.timedelta, td)
                            if pd.notnull(td)
                            else False
                        )
                        varcat.esmcat._df = varcat.df[
                            same_frq | (lower_frq & allow_resampling)
                        ]

                        # For each dataset (id * xrfreq * processing_level * domain * variable),
                        # make sure that file availability covers the requested time periods
                        if periods is not None and len(varcat) > 0:
                            valid_tp = []
                            for var, group in varcat.df.groupby(
                                varcat.esmcat.aggregation_control.groupby_attrs
                                + ["variable"],
                                observed=True,
                            ):
                                valid_tp.append(
                                    subset_file_coverage(
                                        group, periods, **coverage_kwargs
                                    )
                                )  # If valid, this returns the subset of files that cover the time period
                            varcat.esmcat._df = pd.concat(valid_tp)

                        # We now select the coarsest timedelta for each variable
                        # We need to re-iterate over variables in case we used the registry
                        # (and thus there are multiple variables in varcat)
                        rows = []
                        for var, group in varcat.df.groupby("variable"):
                            rows.append(group[group.timedelta == group.timedelta.max()])
                        if rows:
                            # check if the requested variable exists and if so, remove DeriveVariable references
                            v_list = [
                                rows[i]["variable"].iloc[0] for i in range(len(rows))
                            ]
                            v_list_check = [
                                var_id in v_list[i] for i in range(len(v_list))
                            ]  # necessary in case a file has multiple variables
                            if any(v_list_check):
                                rows = [rows[v_list_check.index(True)]]
                                varcat.derivedcat = DerivedVariableRegistry()
                            varcat.esmcat._df = pd.concat(rows, ignore_index=True)
                        else:
                            varcat.esmcat._df = pd.DataFrame()

                    if varcat.df.empty:
                        msg = f"Dataset {sim_id} doesn't have all needed variables (missing at least {var_id})."
                        logger.debug(msg)
                        break
                    if "timedelta" in varcat.df.columns:
                        varcat.df.drop(columns=["timedelta"], inplace=True)
                    varcat._requested_variable_freqs = [xrfreq]
                    varcats.append(varcat)

                else:
                    continue
                break
            else:
                catalogs[sim_id] = concat_data_catalogs(*varcats)
                if periods is not None:
                    catalogs[sim_id]._requested_periods = periods

    if len(catalogs) > 0:
        msg = f"Found {len(catalogs)} with all variables requested and corresponding to the criteria."
        logger.info(msg)
    else:
        logger.warning("Found no match corresponding to the search criteria.")

    if restrict_resolution is not None and len(catalogs) > 0:
        catalogs = _restrict_by_resolution(catalogs, restrict_resolution, id_columns)

    if restrict_members is not None and len(catalogs) > 0:
        catalogs = _restrict_multimembers(catalogs, restrict_members, id_columns)

    return catalogs


@parse_config
def get_warming_level(
    *args, **kwargs
) -> xr.Dataset | xr.DataArray | dict | pd.Series | pd.DataFrame | str | list:
    """
    Deprecated. Use get_period_from_warming_level instead.

    Parameters
    ----------
    args: list
        Arguments to pass to get_period_from_warming_level
    kwargs: dict
        Keyword arguments to pass to get_period_from_warming_level

    Returns
    -------
    xr.Dataset or xr.DataArray or dict or list or str
        Output of get_period_from_warming_level

    """
    kwargs = kwargs.copy()
    kwargs["return_central_year"] = not kwargs.get("return_horizon", True)
    kwargs.pop("return_horizon", None)
    warnings.warn(
        "get_warming_level has been deprecated. Use get_period_from_warming_level instead.",
        FutureWarning,
    )
    return get_period_from_warming_level(*args, **kwargs)


@parse_config
def get_period_from_warming_level(  # noqa: C901
    realization: (
        xr.Dataset | xr.DataArray | dict | pd.Series | pd.DataFrame | str | list
    ),
    wl: float,
    *,
    window: int = 20,
    tas_baseline_period: Sequence[str] | None = None,
    ignore_member: bool = False,
    tas_src: str | os.PathLike | None = None,
    return_central_year: bool = False,
) -> xr.Dataset | xr.DataArray | dict | pd.Series | pd.DataFrame | str | list:
    """
    Use the IPCC Atlas method to return the window of time
    over which the requested level of global warming is first reached.

    Parameters
    ----------
    realization : xr.Dataset, xr.DataArray, dict, str, Series or sequence of those
       Model to be evaluated. Needs the four fields mip_era, source, experiment and member,
       as a dict or in a Dataset's attributes.
       Strings should follow this formatting: {mip_era}_{source}_{experiment}_{member}.
       Lists of dicts, strings or Datasets are also accepted, in which case the output will be a dict.
       Regex wildcards (.*) are accepted, but may lead to unexpected results.
       Datasets should include the catalogue attributes (starting by "cat:") required to create such a string:
       'cat:mip_era', 'cat:experiment', 'cat:member',
       and either 'cat:source' for global models or 'cat:driving_model' for regional models.
       e.g. 'CMIP5_CanESM2_rcp85_r1i1p1'
    wl : float
       Warming level.
       e.g. 2 for a global warming level of +2 degree Celsius above the mean temperature of the `tas_baseline_period`.
    window : int
       Size of the rolling window in years over which to compute the warming level.
    tas_baseline_period : list, optional
       [start, end] of the base period. The warming is calculated with respect to it. The default is ["1850", "1900"].
    ignore_member : bool
       Decides whether to ignore the member when searching for the model run in tas_csv.
    tas_src : str, optional
       Path to a netCDF of annual global mean temperature (tas) with an annual "time" dimension
       and a "simulation" dimension with the following coordinates: "mip_era", "source", "experiment" and "member".
       If None, it will default to data/IPCC_annual_global_tas.nc which was built from
       the IPCC atlas data from  Iturbide et al., 2020 (https://doi.org/10.5194/essd-12-2959-2020)
       and extra data for missing CMIP6 models and pilot models of CRCM5 and ClimEx.
    return_central_year: bool
        If True, the output will be a string representing the middle of the period, using IPCC conventions in the case of an even window (y-9, y+10).
        If False (default), the output will be a list following the format ['start_yr', 'end_yr']

    Returns
    -------
    pd.Series or xr.DataArray or dict or list or str
        If `realization` is not a sequence, the output will follow the format indicated by `return_central_year`.
        If `realization` is a sequence, the output will be of the same type,
        with values following the format indicated by `return_central_year`.
    """
    tas_src = tas_src or Path(__file__).parent / "data" / "IPCC_annual_global_tas.nc"
    tas_baseline_period = standardize_periods(
        tas_baseline_period or ["1850", "1900"], multiple=False
    )

    if (window % 2) not in {0, 1}:
        raise ValueError(f"window should be an integer, received {type(window)}")

    FIELDS = ["mip_era", "source", "experiment", "member"]
    info_models = _wl_prep_infomodels(realization, ignore_member, FIELDS)

    # open nc
    tas = xr.open_dataset(tas_src).tas

    def _get_warming_level(model):
        tas_sel = _wl_find_column(tas, model)
        if tas_sel is None:
            return None if return_central_year else [None, None]

        selected = "_".join([tas_sel[c].item() for c in FIELDS])
        msg = (
            f"Computing warming level +{wl}°C for {model} from simulation: {selected}."
        )
        logger.debug(msg)

        # compute reference temperature for the warming and difference from reference
        yearly_diff = tas_sel - tas_sel.sel(time=slice(*tas_baseline_period)).mean()

        # get the start and end date of the window when the warming level is first reached
        rolling_diff = yearly_diff.rolling(
            time=window, min_periods=window, center=True
        ).mean()
        # shift(-1) is needed to reproduce IPCC results.
        # rolling defines the window as [n-10,n+9], but the the IPCC defines it as [n-9,n+10], where n is the center year.
        if window % 2 == 0:  # Even window
            rolling_diff = rolling_diff.shift(time=-1)

        yrs = rolling_diff.where(rolling_diff >= wl, drop=True)
        if yrs.size == 0:
            msg = (
                f"Global warming level of +{wl}C is not reached by the last year "
                f"({tas.time[-1].dt.year.item()}) of the provided 'tas_src' database for {selected}."
            )
            logger.info(msg)
            return None if return_central_year else [None, None]

        yr = yrs.isel(time=0).time.dt.year.item()
        start_yr = int(yr - window / 2 + 1)
        end_yr = int(yr + window / 2)
        return (
            str(yr)
            if return_central_year
            else standardize_periods([start_yr, end_yr], multiple=False)
        )

    out = list(map(_get_warming_level, info_models))
    if isinstance(realization, pd.DataFrame):
        return pd.Series(out, index=realization.index)
    if isinstance(realization, xr.DataArray):
        if return_central_year is False:
            return xr.DataArray(
                out, dims=(realization.dims[0], "wl_bounds"), coords=realization.coords
            )
        return xr.DataArray(out, dims=(realization.dims[0],), coords=realization.coords)

    if len(out) == 1:
        return out[0]
    return out


@parse_config
def get_warming_level_from_period(
    realization: (
        xr.Dataset | xr.DataArray | dict | pd.Series | pd.DataFrame | str | list
    ),
    period: list[str],
    *,
    tas_baseline_period: Sequence[str] | None = None,
    ignore_member: bool = False,
    tas_src: str | os.PathLike | None = None,
) -> xr.Dataset | xr.DataArray | dict | pd.Series | pd.DataFrame | float | list:
    """
    Return the warming level reached in the given period.

    Parameters
    ----------
    realization : xr.Dataset, xr.DataArray, dict, str, Series or sequence of those
       Model to be evaluated. Needs the four fields mip_era, source, experiment and member,
       as a dict or in a Dataset's attributes.
       Strings should follow this formatting: {mip_era}_{source}_{experiment}_{member}.
       Lists of dicts, strings or Datasets are also accepted, in which case the output will be a dict.
       Regex wildcards (.*) are accepted, but may lead to unexpected results.
       Datasets should include the catalogue attributes (starting by "cat:") required to create such a string:
       'cat:mip_era', 'cat:experiment', 'cat:member',
       and either 'cat:source' for global models or 'cat:driving_model' for regional models.
       e.g. 'CMIP5_CanESM2_rcp85_r1i1p1'
    period : list of str
       [start, end] of the period for which to compute the warming level.
    tas_baseline_period : list, optional
       [start, end] of the base period. The warming is calculated with respect to it. The default is ["1850", "1900"].
    ignore_member : bool
       Decides whether to ignore the member when searching for the model run in tas_csv.
    tas_src : str, optional
       Path to a netCDF of annual global mean temperature (tas) with an annual "time" dimension
       and a "simulation" dimension with the following coordinates: "mip_era", "source", "experiment" and "member".
       If None, it will default to data/IPCC_annual_global_tas.nc which was built from
       the IPCC atlas data from  Iturbide et al., 2020 (https://doi.org/10.5194/essd-12-2959-2020)
       and extra data for missing CMIP6 models and pilot models of CRCM5 and ClimEx.

    Returns
    -------
    pd.Series or xr.DataArray or float
        If `realization` is not a sequence, the output will be a float.
        If `realization` is a sequence, the output will be of the same type.
    """
    tas_src = tas_src or Path(__file__).parent / "data" / "IPCC_annual_global_tas.nc"
    tas_baseline_period = standardize_periods(
        tas_baseline_period or ["1850", "1900"], multiple=False
    )
    period = standardize_periods(period, multiple=False)

    FIELDS = ["mip_era", "source", "experiment", "member"]
    info_models = _wl_prep_infomodels(realization, ignore_member, FIELDS)

    # open nc
    tas = xr.open_dataset(tas_src).tas

    def _get_warming_level(model):
        tas_sel = _wl_find_column(tas, model)
        if tas_sel is None:
            return None

        selected = "_".join([tas_sel[c].item() for c in FIELDS])
        msg = f"Computing warming level during {period} for {model} from simulation: {selected}."
        logger.debug(msg)

        if not all(
            yr in tas_sel.time.dt.year
            for yr in np.arange(int(period[0]), int(period[1]) + 1)
        ):
            raise ValueError(
                f"Period {period} is not fully covered by the provided 'tas_src' database for {selected}."
            )

        # compute reference temperature for the warming and difference from reference
        wl = (
            tas_sel.sel(time=slice(*period)).mean()
            - tas_sel.sel(time=slice(*tas_baseline_period)).mean()
        )

        return wl.item()

    out = list(map(_get_warming_level, info_models))
    if isinstance(realization, pd.DataFrame):
        return pd.Series(out, index=realization.index)
    if isinstance(realization, xr.DataArray):
        return xr.DataArray(out, dims=(realization.dims[0],), coords=realization.coords)

    if len(out) == 1:
        return out[0]
    return out


def _wl_prep_infomodels(realization, ignore_member, fields):
    if isinstance(realization, xr.Dataset | str | dict | pd.Series):
        reals = [realization]
    elif isinstance(realization, pd.DataFrame):
        reals = (row for i, row in realization.iterrows())
    elif isinstance(realization, xr.DataArray):
        reals = realization.values
    else:
        reals = realization

    info_models = []
    for real in reals:
        info = {}
        if isinstance(real, xr.Dataset):
            attrs = get_cat_attrs(real)
            # get info on ds
            if attrs.get("driving_model") is None:
                info["source"] = attrs["source"]
            else:
                info["source"] = attrs["driving_model"]
            info["experiment"] = attrs["experiment"]
            info["member"] = ".*" if ignore_member else attrs["member"]
            info["mip_era"] = attrs["mip_era"]
        elif isinstance(real, str):
            (
                info["mip_era"],
                info["source"],
                info["experiment"],
                info["member"],
            ) = real.split("_")
            if ignore_member:
                info["member"] = ".*"
        # Dict or Series (DataFrame row)
        elif hasattr(real, "keys") and set(real.keys()).issuperset(
            (set(fields) - {"member"}) if ignore_member else fields
        ):
            info = real
            if info.get("driving_model") is not None:
                info["source"] = info["driving_model"]
            if ignore_member:
                info["member"] = ".*"
        else:
            raise ValueError(
                f"'realization' must be a Dataset, dict, string or list. Received {type(real)}."
            )
        info_models.append(info)

    return info_models


def _wl_find_column(tas, model):
    # Choose column based on ds cat attrs, +'$' to ensure a full match (matches end-of-string)
    mip = tas.mip_era.str.match(model["mip_era"] + "$")
    src = tas.source.str.match(model["source"] + "$")
    if not src.any():
        # Maybe it's an RCM, then requested source may contain the institute
        src = xr.apply_ufunc(model["source"].endswith, tas.source, vectorize=True)
    exp = tas.experiment.str.match(model["experiment"] + "$")
    mem = tas.member.str.match(model["member"] + "$")

    candidates = mip & src & exp & mem
    if not candidates.any():
        warnings.warn(
            f"No simulation fit the attributes of the input dataset ({model})."
        )
        return None

    if candidates.sum() > 1:
        logger.info(
            "More than one simulation of the database fits the dataset metadata. Choosing the first one."
        )
    tas_sel = tas.isel(simulation=candidates.argmax("simulation"))
    return tas_sel


@parse_config
def subset_warming_level(
    ds: xr.Dataset,
    wl: float | Sequence[float],
    to_level: str = "warminglevel-{wl}vs{period0}-{period1}",
    wl_dim: str | bool = "+{wl}Cvs{period0}-{period1}",
    **kwargs,
) -> xr.Dataset | None:
    r"""
    Subsets the input dataset with only the window of time over which the requested level of global warming
    is first reached, using the IPCC Atlas method.
    A warming level is considered reached only if the full `window` years are available in the dataset.

    Parameters
    ----------
    ds : xr.Dataset
       Input dataset.
       The dataset should include attributes to help recognize it and find its
       warming levels - 'cat:mip_era', 'cat:experiment', 'cat:member', and either
       'cat:source' for global models or 'cat:driving_institution' (optional) + 'cat:driving_model' for regional models.
       Or , it should include a `realization` dimension constructed as "{mip_era}_{source or driving_model}_{experiment}_{member}"
       for vectorized subsetting. Vectorized subsetting is currently only implemented for annual data.
    wl : float or sequence of floats
       Warming level.
       e.g. 2 for a global warming level of +2 degree Celsius above the mean temperature of the `tas_baseline_period`.
       Multiple levels can be passed, in which case using "{wl}" in  `to_level` and `wl_dim` is not recommended.
       Multiple levels are currently only implemented for annual data.
    to_level :
       The processing level to assign to the output.
       Use "{wl}", "{period0}" and "{period1}" in the string to dynamically include
       `wl`, 'tas_baseline_period[0]' and 'tas_baseline_period[1]'.
    wl_dim : str or boolean, optional
       The value to use to fill the new `warminglevel` dimension.
       Use "{wl}", "{period0}" and "{period1}" in the string to dynamically include
       `wl`, 'tas_baseline_period[0]' and 'tas_baseline_period[1]'.
       If None, no new dimensions will be added, invalid if `wl` is a sequence.
       If True, the dimension will include `wl` as numbers and units of "degC".
    \*\*kwargs :
        Instructions on how to search for warming levels, passed to :py:func:`get_period_from_warming_level`.

    Returns
    -------
    xr.Dataset or None
        Warming level dataset, or None if `ds` can't be subsetted for the requested warming level.
        The dataset will have a new dimension `warminglevel` with `wl_dim` as coordinates.
        If `wl` was a list or if ds had a "realization" dim, the "time" axis is replaced
        by a fake time starting in 1000-01-01 and with a length of `window` years.
        Start and end years of the subsets are bound in the new coordinate "warminglevel_bounds".
    """
    tas_baseline_period = standardize_periods(
        kwargs.get("tas_baseline_period", ["1850", "1900"]), multiple=False
    )
    window = kwargs.get("window", 20)

    # If wl was originally a list, this function is called a 2nd time with a generated fake_time
    fake_time = kwargs.pop("_fake_time", None)
    # Fake time generation is needed : real is a dim or multiple levels
    if (
        fake_time is None
        and not isinstance(wl, int | float)
        or "realization" in ds.dims
    ):
        freq = xr.infer_freq(ds.time)
        # FIXME: This is because I couldn't think of an elegant way to generate a fake_time otherwise.
        if not compare_offsets(freq, "==", "YS"):
            raise NotImplementedError(
                "Passing multiple warming levels or vectorizing subsetting along the 'realization' dim is currently not supported for non-annual data"
            )
        fake_time = xr.date_range(
            "1000-01-01", periods=window, freq=freq, calendar=ds.time.dt.calendar
        )

    # If we got a wl sequence, call ourself multiple times and concatenate
    if not isinstance(wl, int | float):
        if not wl_dim or (isinstance(wl_dim, str) and "{wl}" not in wl_dim):
            raise ValueError(
                "`wl_dim` must be True or a template string including '{wl}' if multiple levels are passed."
            )
        ds_wl = xr.concat(
            [
                subset_warming_level(
                    ds,
                    wli,
                    to_level=to_level,
                    wl_dim=wl_dim,
                    _fake_time=fake_time,
                    **kwargs,
                )
                for wli in wl
            ],
            "warminglevel",
        )
        return ds_wl

    # Creating the warminglevel coordinate
    if isinstance(wl_dim, str):  # a non-empty string
        wl_crd = xr.DataArray(
            [
                wl_dim.format(
                    wl=wl,
                    period0=tas_baseline_period[0],
                    period1=tas_baseline_period[1],
                )
            ],
            dims=("warminglevel",),
            name="warminglevel",
        )
    else:
        wl_crd = xr.DataArray(
            [wl], dims=("warminglevel",), name="warminglevel", attrs={"units": "degC"}
        )

    # For generating the bounds coord
    date_cls = xc.core.calendar.datetime_classes[ds.time.dt.calendar]
    if "realization" in ds.dims:
        # Vectorized subset
        bounds = get_period_from_warming_level(
            ds.realization, wl, return_central_year=False, **kwargs
        )
        reals = []
        for real in bounds.realization.values:
            start, end = bounds.sel(realization=real).values
            data = ds.sel(realization=[real], time=slice(start, end))
            wl_not_reached = (
                (start is None)
                or (data.time.size == 0)
                or ((data.time.dt.year[-1] - data.time.dt.year[0] + 1) != window)
            )
            if not wl_not_reached:
                bnds_crd = [
                    date_cls(int(start), 1, 1),
                    date_cls(int(end) + 1, 1, 1) - datetime.timedelta(seconds=1),
                ]
            else:
                # In the case of not reaching the WL, data might be too short
                # We create it again with the proper length
                data = (
                    ds.sel(realization=[real]).isel(time=slice(0, fake_time.size))
                    * np.nan
                )
                bnds_crd = [np.nan, np.nan]
            reals.append(
                data.expand_dims(warminglevel=wl_crd).assign_coords(
                    time=fake_time[: data.time.size],
                    warminglevel_bounds=(
                        ("realization", "warminglevel", "wl_bounds"),
                        [[bnds_crd]],
                    ),
                )
            )
        ds_wl = xr.concat(reals, "realization")
    else:
        # Scalar subset, single level
        start_yr, end_yr = get_period_from_warming_level(
            ds, wl=wl, return_central_year=False, **kwargs
        )
        # cut the window selected above and expand dims with wl_crd
        ds_wl = ds.sel(time=slice(start_yr, end_yr))
        wl_not_reached = (
            (start_yr is None)
            or (ds_wl.time.size == 0)
            or ((ds_wl.time.dt.year[-1] - ds_wl.time.dt.year[0] + 1) != window)
        )
        if fake_time is None:
            # WL not reached, not in ds, or not fully contained in ds.time
            if wl_not_reached:
                return None
            ds_wl = ds_wl.expand_dims(warminglevel=wl_crd)
        else:
            # WL not reached, not in ds, or not fully contained in ds.time
            if wl_not_reached:
                ds_wl = ds.isel(time=slice(0, fake_time.size)) * np.nan
                wlbnds = (("warminglevel", "wl_bounds"), [[np.nan, np.nan]])
            else:
                wlbnds = (
                    ("warminglevel", "wl_bounds"),
                    [
                        [
                            date_cls(int(start_yr), 1, 1),
                            date_cls(int(end_yr) + 1, 1, 1)
                            - datetime.timedelta(seconds=1),
                        ]
                    ],
                )
            # We are in an iteration over multiple levels, put the fake time axis, but remember bounds
            ds_wl = ds_wl.expand_dims(warminglevel=wl_crd).assign_coords(
                time=fake_time[: ds_wl.time.size],
                warminglevel_bounds=wlbnds,
            )

    if to_level is not None:
        ds_wl.attrs["cat:processing_level"] = to_level.format(
            wl=wl,
            period0=tas_baseline_period[0],
            period1=tas_baseline_period[1],
        )

    if not wl_dim:
        ds_wl = ds_wl.squeeze("warminglevel", drop=True)
    else:
        ds_wl.warminglevel.attrs.update(
            baseline=f"{tas_baseline_period[0]}-{tas_baseline_period[1]}",
            long_name=f"Warming level for {window}-year periods since {tas_baseline_period[0]}-{tas_baseline_period[1]}",
        )
    return ds_wl


def _dispatch_historical_to_future(
    catalog: DataCatalog, id_columns: list[str] | None = None
) -> DataCatalog:
    """Update a DataCatalog by recopying each "historical" entry to its corresponding future experiments.

    For examples, if an historical entry has corresponding "ssp245" and "ssp585" entries,
    then it is copied twice, with its "experiment" field modified accordingly.
    The original "historical" entry is removed. This way, a subsequent search of the catalog
    with "experiment='ssp245'" includes the _historical_ assets (with no apparent distinction).

    "Historical" assets that did not find a match are removed from the output catalog.

    Parameters
    ----------
    catalog : DataCatalog
        Catalog to be evaluated.
    id_columns : list of str, optional
        List of columns to be used to identify unique simulations.
        If None, defaults to ID_COLUMNS.

    Returns
    -------
    DataCatalog
        Catalog with the historical entries duplicated and modified to match the future experiments.
    """
    expcols = [
        "experiment",
        "id",
        "variable",
        "xrfreq",
        "date_start",
        "date_end",
        "path",
        "format",
        "activity",
    ]  # These columns can differentiate the historical from the future.
    sim_id_no_exp = list(
        filter(
            lambda c: c not in expcols,
            set(id_columns or ID_COLUMNS).intersection(catalog.df.columns),
        )
    )
    # For each unique element from this column list, the historical shouldn't be distinct
    # from the future scenario, allowing us to match them.
    # "Same hist member" as in "each future realization stems from the same historical member"

    df = catalog.df.copy()
    df["same_hist_member"] = df[sim_id_no_exp].apply(
        lambda row: "_".join(row.values.astype(str)), axis=1
    )

    new_lines = []
    for group in df.same_hist_member.unique():
        sdf = df[df.same_hist_member == group]
        hist = sdf[sdf.experiment == "historical"]
        if hist.empty:
            continue
        if pd.isna(sdf.activity).any():
            warnings.warn(
                f"np.NaN was found in the activity column of {group}. The rows with np.NaN activity will be skipped."
                "If you want them to be included in the historical and future matching, "
                "please put a valid activity (https://xscen.readthedocs.io/en/latest/columns.html)."
                "For example, xscen expects experiment `historical` to have `CMIP` activity "
                "and experiments `sspXYZ` to have `ScenarioMIP` activity. "
            )
        for activity_id in set(sdf.activity) - {"HighResMip", np.nan}:
            sub_sdf = sdf[sdf.activity == activity_id]
            for exp_id in set(sub_sdf.experiment) - {"historical", "piControl", np.nan}:
                exp_hist = hist.copy()
                exp_hist["experiment"] = exp_id
                exp_hist["activity"] = activity_id
                exp_hist["same_hist_member"] = group
                sim_ids = sub_sdf[sub_sdf.experiment == exp_id].id.unique()
                if len(sim_ids) > 1:
                    raise ValueError(
                        f"Got multiple dataset ids where we expected only one... : {sim_ids}"
                    )
                exp_hist["id"] = sim_ids[0]

                # Remove fixed fields that already exist in the future experiment
                dupes = pd.concat(
                    [
                        exp_hist.loc[exp_hist["frequency"] == "fx"],
                        sub_sdf.loc[
                            (sub_sdf["frequency"] == "fx")
                            & (sub_sdf["experiment"] == exp_id)
                        ],
                    ]
                ).duplicated(
                    subset=[
                        "same_hist_member",
                        "variable",
                        "frequency",
                        "experiment",
                        "activity",
                    ],
                    keep=False,
                )
                dupes = dupes[dupes]  # Only keep the duplicated rows
                exp_hist = exp_hist.loc[exp_hist.index.difference(dupes.index)]

                new_lines.append(exp_hist)

    df = pd.concat([df] + new_lines, ignore_index=True).drop(
        columns=["same_hist_member"]
    )
    df = df[df.experiment != "historical"].reset_index(drop=True)
    return DataCatalog(
        {"esmcat": catalog.esmcat.model_dump(), "df": df},
        registry=catalog.derivedcat,
        drop_duplicates=False,
    )


def _restrict_by_resolution(
    catalogs: dict, restrictions: str, id_columns: list[str] | None = None
) -> dict:
    """Update the results from search_data_catalogs by removing simulations with multiple resolutions available.

    Parameters
    ----------
    catalogs : dict
        Dictionary of DataCatalogs to be evaluated.
    restrictions : str
        Either 'finest' or 'coarsest'.
    id_columns : list of str, optional
        List of columns to be used to identify unique simulations.
        If None, defaults to ID_COLUMNS.

    Returns
    -------
    dict
        Catalogs with duplicate simulations removed according to the resolution restrictions.

    Notes
    -----
    Currently supports:
        CMIP5 and CMIP6 (in that order, giving the highest priority to elements on the left):
            [gn, gn{a/g}, gr, gr{0-9}{a/g}, global, gnz, gr{0-9}z, gm]
        CORDEX:
            [DOM-{resolution}, DOM-{resolution}i]
    """
    df = pd.concat([catalogs[s].df for s in catalogs.keys()])
    # remove the domain from the group_by
    df["id_nodom"] = df[
        list(
            set(id_columns or ID_COLUMNS)
            .intersection(df.columns)
            .difference(["domain"])
        )
    ].apply(lambda row: "_".join(map(str, filter(pd.notna, row.values))), axis=1)
    for i in pd.unique(df["id_nodom"]):
        df_sim = df[df["id_nodom"] == i]
        domains = pd.unique(df_sim["domain"])

        if len(domains) > 1:
            msg = f"Dataset {i} appears to have multiple resolutions."
            logger.info(msg)

            # For CMIP, the order is dictated by a list of grid labels
            if "MIP" in pd.unique(df_sim["activity"])[0]:
                order = np.array([])
                for d in domains:
                    match = [
                        CV.infer_resolution("CMIP").index(r)
                        for r in CV.infer_resolution("CMIP")
                        if re.match(pattern=r, string=d)
                    ]
                    if len(match) != 1:
                        raise ValueError(f"'{d}' matches no known CMIP domain.")
                    order = np.append(order, match[0])

                if restrictions == "finest":
                    chosen = [
                        np.sort(
                            np.array(domains)[
                                np.where(order == np.array(order).min())[0]
                            ]
                        )[0]
                    ]
                elif restrictions == "coarsest":
                    chosen = [
                        np.sort(
                            np.array(domains)[
                                np.where(order == np.array(order).max())[0]
                            ]
                        )[-1]
                    ]
                else:
                    raise ValueError(
                        "'restrict_resolution' should be 'finest' or 'coarsest'"
                    )

            # Note: For CORDEX, the order is dictated by both the grid label
            # and the resolution itself as well as the domain name
            elif pd.unique(df_sim["activity"])[0] == "CORDEX":
                # Unique CORDEX domains
                cordex_doms = list(
                    pd.unique(pd.Series([d.split("-")[0] for d in domains]))
                )
                chosen = []

                for d in cordex_doms:
                    sub = [doms for doms in domains if d in doms]
                    order = [
                        int(re.split("^([A-Z]{3})-([0-9]{2})([i]{0,1})$", s)[2])
                        for s in sub
                    ]

                    if restrictions == "finest":
                        chosen.extend(
                            [
                                np.sort(
                                    np.array(sub)[
                                        np.where(order == np.array(order).min())[0]
                                    ]
                                )[0]
                            ]
                        )
                    elif restrictions == "coarsest":
                        chosen.extend(
                            [
                                np.sort(
                                    np.array(sub)[
                                        np.where(order == np.array(order).max())[0]
                                    ]
                                )[0]
                            ]
                        )
                    else:
                        raise ValueError(
                            "'restrict_resolution' should be 'finest' or 'coarsest'"
                        )

            else:
                msg = f"Dataset {i} seems to have multiple resolutions, but its activity is not yet recognized or supported."
                logger.warning(msg)
                chosen = list(domains)
                pass

            to_remove = pd.unique(
                df_sim[
                    df_sim["domain"].isin(
                        list(set(pd.unique(df_sim["domain"])).difference(chosen))
                    )
                ]["id"]
            )

            for k in to_remove:
                msg = f"Removing {k} from the results."
                logger.info(msg)
                catalogs.pop(k)

    return catalogs


def _restrict_multimembers(
    catalogs: dict, restrictions: dict, id_columns: list[str] | None = None
):
    """Update the results from search_data_catalogs by removing simulations with multiple members available.

    Uses regex to try and adequately detect and order the member's identification number, but only tested for 'r-i-p'.

    Parameters
    ----------
    catalogs : dict
        Dictionary of DataCatalogs to be evaluated.
    restrictions : dict
        Dictionary of restrictions to be applied. Currently only supports {'ordered': int}.
    id_columns : list of str, optional
        List of columns to be used to identify unique simulations.
        If None, defaults to ID_COLUMNS.

    Returns
    -------
    dict
        Catalogs where simulations with multiple members have been restricted to the requested maximum number.
    """
    df = pd.concat([catalogs[s].df for s in catalogs.keys()])
    # remove the member from the group_by
    df["id_nomem"] = df[
        list(
            set(id_columns or ID_COLUMNS)
            .intersection(df.columns)
            .difference(["member"])
        )
    ].apply(lambda row: "_".join(map(str, filter(pd.notna, row.values))), axis=1)

    for i in pd.unique(df["id_nomem"]):
        df_sim = df[df["id_nomem"] == i]
        members = pd.unique(df_sim["member"])

        if len(members) > 1:
            msg = f"Dataset {i} has {len(members)} valid members. Restricting as per requested."
            logger.info(msg)

            if "ordered" in restrictions:
                members = natural_sort(members)[0 : restrictions["ordered"]]
            else:
                raise NotImplementedError(
                    "Subsetting multiple members currently only supports 'ordered'."
                )

            to_remove = pd.unique(
                df_sim[
                    df_sim["member"].isin(
                        list(set(pd.unique(df_sim["member"])).difference(members))
                    )
                ]["id"]
            )

            for k in to_remove:
                msg = f"Removing {k} from the results."
                logger.info(msg)
                catalogs.pop(k)

    return catalogs


def _restrict_wl(df: pd.DataFrame, restrictions: dict):
    """Update the results from search_data_catalogs according to warming level restrictions.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to be evaluated.
    restrictions : dict
        Dictionary of restrictions to be applied. Entries are passed to get_period_from_warming_level.
        If 'wl' is present, the warming level csv will be used to remove simulations that do not reach the requested warming level.
        Otherwise, the warming level csv will be used to remove simulations that are not available in it.

    Returns
    -------
    df :
        Updated DataFrame.
    """
    restrictions.setdefault("wl", 0)
    to_keep = get_period_from_warming_level(
        df, return_central_year=True, **restrictions
    ).notnull()
    removed = pd.unique(df[~to_keep]["id"])
    df = df[to_keep]
    msg = f"Removing the following datasets because of the restriction for warming levels: {list(removed)}"
    logger.info(msg)
    return df



"""A climate change scenario-building analysis framework, built with xclim/xarray."""

###################################################################################
# Apache Software License 2.0
#
# Copyright (c) 2024, Gabriel Rondeau-Genesse, Pascal Bourgault, Juliette Lavoie, Trevor James Smith
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###################################################################################

import warnings

# Import the submodules
from . import (
    aggregate,
    biasadjust,
    catalog,
    catutils,
    config,
    diagnostics,
    ensembles,
    extract,
    indicators,
    io,
    regrid,
    scripting,
    spatial,
    testing,
    utils,
)

# Import top-level functions
from .aggregate import *
from .biasadjust import *
from .catalog import DataCatalog, ProjectCatalog
from .catutils import build_path, parse_directory
from .config import CONFIG, load_config
from .diagnostics import properties_and_measures
from .ensembles import *
from .extract import (
    extract_dataset,
    get_period_from_warming_level,
    get_warming_level,
    get_warming_level_from_period,
    search_data_catalogs,
    subset_warming_level,
)
from .indicators import compute_indicators
from .io import save_to_netcdf, save_to_table, save_to_zarr
from .regrid import *
from .scripting import (
    TimeoutException,
    measure_time,
    move_and_delete,
    save_and_update,
    send_mail,
    send_mail_on_exit,
    timeout,
)
from .utils import clean_up

__author__ = """Gabriel Rondeau-Genesse"""
__email__ = "rondeau-genesse.gabriel@ouranos.ca"
__version__ = "0.12.0"


# FIXME: file and line are unused
def warning_on_one_line(
    message: str,
    category: Warning,
    filename: str,
    lineno: int,
    file=None,  # noqa: F841
    line=None,  # noqa: F841
):
    """
    Monkeypatch Reformat warning so that `warnings.warn` doesn't mention itself.

    Parameters
    ----------
    message : str
        The warning message.
    category : Warning
        The warning category.
    filename : str
        The filename where the warning was raised.
    lineno : int
        The line number where the warning was raised.
    file : file
        The file where the warning was raised.
    line : str
        The line where the warning was raised.

    Returns
    -------
    str
        The reformatted warning message.
    """
    return f"{filename}:{lineno}: {category.__name__}: {message}\n"


warnings.formatwarning = warning_on_one_line

# FIXME: This is a temporary fix for the FutureWarning spam from intake-esm.
# Print FutureWarnings from intake-esm only once
warnings.filterwarnings(
    "ignore",
    category=FutureWarning,
    module="intake_esm",
    message="The default of observed=False is deprecated and will be changed to True in a future version of pandas. "
    "Pass observed=False to retain current behavior or observed=True to adopt the future default "
    "and silence this warning.",
)



import warnings

import numpy as np
import pytest
import xclim
from conftest import notebooks
from xclim.testing.helpers import test_timeseries as timeseries

import xscen as xs


class TestComputeIndicators:
    yaml_file = notebooks / "samples" / "indicators.yml"
    ds = timeseries(
        np.ones(365 * 3), variable="tas", start="2001-01-01", freq="D", as_dataset=True
    )

    @pytest.mark.parametrize("reload", [True, False])
    def test_reload_module(self, reload):
        module = xs.indicators.load_xclim_module(self.yaml_file)
        assert all(hasattr(module, ind) for ind in ["growing_degree_days", "tg_min"])

        # Record warnings without failing the test if no warnings are raised.
        with warnings.catch_warnings(record=True) as record:
            warnings.simplefilter("always")
            xs.indicators.load_xclim_module(
                notebooks / "samples" / "indicators", reload=reload
            )
        assert len(
            [
                r
                for r in record
                if "already exists and will be overwritten." in str(r.message)
            ]
        ) == (2 if reload else 0)

    @pytest.mark.parametrize("input", ["module", "iter"])
    def test_input_types(self, input):
        module = xs.indicators.load_xclim_module(self.yaml_file)
        ind_dict1 = xs.compute_indicators(
            self.ds,
            indicators=module if input == "module" else module.iter_indicators(),
        )
        ind_dict2 = xs.compute_indicators(self.ds, indicators=self.yaml_file)
        assert all(ind_dict1[k].equals(ind_dict2[k]) for k in ind_dict1)

    @pytest.mark.parametrize("to_level", [None, "test"])
    def test_level(self, to_level):
        if to_level is None:
            ind_dict = xs.compute_indicators(self.ds, indicators=self.yaml_file)
            assert "indicators" in ind_dict["YS-JAN"].attrs["cat:processing_level"]
        else:
            ind_dict = xs.compute_indicators(
                self.ds, indicators=self.yaml_file, to_level=to_level
            )
            assert to_level in ind_dict["YS-JAN"].attrs["cat:processing_level"]

    # Periods needed to cover both branches of the if/else in compute_indicators, so no need to test separately.
    @pytest.mark.parametrize(
        "periods", [None, [["2001", "2001"], ["2003", "2004"], ["2005", "2009"]]]
    )
    def test_output(self, periods):
        values = np.ones(365 * (2 if periods is None else 10))
        ds = timeseries(
            values, variable="tas", start="2001-01-01", freq="D", as_dataset=True
        )
        ds["da"] = ds.tas

        module = xs.indicators.load_xclim_module(self.yaml_file)
        ind_dict = xs.compute_indicators(
            ds,
            indicators=[("fit", xclim.indicators.generic.fit)]
            + [x for x in module.iter_indicators()],
            periods=periods,
        )
        assert all(
            xrfreq in ind_dict[xrfreq].attrs["cat:xrfreq"] for xrfreq in ind_dict.keys()
        )
        assert all(v in ind_dict["YS-JAN"] for v in ["tg_min", "growing_degree_days"])
        assert all(
            v in ind_dict["YS-JAN"].attrs["cat:variable"]
            for v in ["tg_min", "growing_degree_days"]
        )
        if periods is None:
            assert "time" not in ind_dict["fx"].dims
            assert len(ind_dict["YS-JAN"].time) == 2
        else:
            assert len(ind_dict["fx"].time) == len(periods)
            assert len(ind_dict["YS-JAN"].time) == 8

    def test_qs_dec(self):
        indicator = xclim.core.indicator.Indicator.from_dict(
            data={"base": "tg_min", "parameters": {"freq": "QS-DEC"}},
            identifier="tg_min_qs",
            module="tests",
        )
        ind_dict = xs.compute_indicators(self.ds, indicators=[("tg_min_qs", indicator)])
        assert "QS-DEC" in ind_dict["QS-DEC"].attrs["cat:xrfreq"]
        assert len(ind_dict["QS-DEC"].time) == 12
        assert ind_dict["QS-DEC"].time[0].dt.strftime("%Y-%m-%d").item() == "2001-03-01"
        assert (
            ind_dict["QS-DEC"].time[-1].dt.strftime("%Y-%m-%d").item() == "2003-12-01"
        )

    # Periods needed to cover both branches of the if/else in compute_indicators, so no need to test separately.
    @pytest.mark.parametrize("periods", [None, [["2001", "2001"], ["2003", "2003"]]])
    def test_multiple_outputs(self, periods):
        values = np.zeros(365 * 3)
        values[150:250] = 100
        values[150 + 365 : 250 + 365] = 100
        values[150 + 365 * 2 : 250 + 365 * 2] = 100
        ds = timeseries(
            values, variable="pr", start="2001-01-01", freq="D", as_dataset=True
        )
        ind_dict = xs.compute_indicators(
            ds,
            indicators=[
                ("precip", xclim.atmos.precip_average),
                ("rs", xclim.atmos.rain_season),
            ],
            periods=periods,
        )

        assert all(
            v in ind_dict["YS-JAN"]
            for v in [
                "prcpavg",
                "rain_season_start",
                "rain_season_end",
                "rain_season_length",
            ]
        )

    @pytest.mark.parametrize("restrict_years", [True, False])
    def test_as_jul(self, restrict_years):
        indicator = xclim.core.indicator.Indicator.from_dict(
            data={"base": "freezing_degree_days", "parameters": {"freq": "YS-JUL"}},
            identifier="degree_days_below_0_annual_start_july",
            module="tests",
        )
        ind_dict = xs.compute_indicators(
            self.ds,
            indicators=[("degree_days_below_0_annual_start_july", indicator)],
            restrict_years=restrict_years,
        )

        assert "YS-JUL" in ind_dict["YS-JUL"].attrs["cat:xrfreq"]
        out = ind_dict["YS-JUL"]
        if restrict_years:
            assert len(out.time) == 3  # same length as input ds
            assert out.time[0].dt.strftime("%Y-%m-%d").item() == "2001-07-01"
            assert out.time[-1].dt.strftime("%Y-%m-%d").item() == "2003-07-01"
        else:
            assert len(out.time) == 4
            assert out.time[0].dt.strftime("%Y-%m-%d").item() == "2000-07-01"
            assert out.time[-1].dt.strftime("%Y-%m-%d").item() == "2003-07-01"

    @pytest.mark.parametrize("indicator_iter", ["list", "tuples", "module"])
    def test_select_inds_for_avail_vars(self, indicator_iter):
        # Test that select_inds_for_avail_vars filters a list of indicators to only those
        # that can be computed with the variables available in a dataset.
        ds = xs.testing.datablock_3d(
            np.ones((365, 3, 3)),
            variable="tas",
            x="lon",
            x_start=-75,
            y="lat",
            y_start=45,
            x_step=1,
            y_step=1,
            start="2001-01-01",
            freq="D",
            units="K",
            as_dataset=True,
        )
        indicators = [
            xclim.core.indicator.Indicator.from_dict(
                data={"base": "tg_min", "parameters": {"freq": "QS-DEC"}},
                identifier="tg_min_qs",
                module="tests",
            ),
            xclim.core.indicator.Indicator.from_dict(
                data={"base": "days_over_precip_thresh", "parameters": {"freq": "MS"}},
                identifier="precip_average_ms",
                module="tests",
            ),
        ]

        # indicators as different types
        module = xclim.core.indicator.build_indicator_module(
            "indicators", {i.base: i for i in indicators}, reload=True
        )
        if indicator_iter == "list":
            inds_for_avail_vars = xs.indicators.select_inds_for_avail_vars(
                ds=ds, indicators=indicators
            )
        elif indicator_iter == "tuples":
            inds_for_avail_vars = xs.indicators.select_inds_for_avail_vars(
                ds=ds, indicators=[(n, i) for n, i in module.iter_indicators()]
            )
        elif indicator_iter == "module":
            inds_for_avail_vars = xs.indicators.select_inds_for_avail_vars(
                ds=ds, indicators=module
            )

        assert len(list(inds_for_avail_vars.iter_indicators())) == 1
        assert [n for n, _ in inds_for_avail_vars.iter_indicators()] == ["tg_min"]
        assert [i for _, i in inds_for_avail_vars.iter_indicators()] == [indicators[0]]
        # no indicators found
        inds_for_avail_vars = xs.indicators.select_inds_for_avail_vars(
            ds=ds, indicators=indicators[1:]
        )
        assert len(list(inds_for_avail_vars.iter_indicators())) == 0
        assert [(n, i) for n, i in inds_for_avail_vars.iter_indicators()] == []


@pytest.mark.parametrize(
    "ind,expvars,expfrq",
    [
        ("wind_vector_from_speed", ["uas", "vas"], "D"),
        ("fit", ["params"], "fx"),
        ("tg_mean", ["tg_mean"], "YS-JAN"),
    ],
)
def test_get_indicator_outputs(ind, expvars, expfrq):
    ind = xclim.core.indicator.registry[ind.upper()].get_instance()
    outvars, outfrq = xs.indicators.get_indicator_outputs(ind, "D")
    assert outvars == expvars
    assert outfrq == expfrq



from datetime import datetime
from pathlib import Path

import cftime
import dask.array
import numpy as np
import pandas as pd
import pytest
import xarray as xr
import xclim as xc
from xclim.testing.helpers import test_timeseries as timeseries

import xscen as xs
from xscen.testing import datablock_3d


class TestLocale:
    def test_update(self):
        ds = timeseries(
            np.tile(np.arange(1, 366), 30),
            variable="tas",
            start="2001-01-01",
            freq="D",
            as_dataset=True,
        )
        indicator = xc.core.indicator.Indicator.from_dict(
            data={"base": "tg_mean"},
            identifier="tg_mean",
            module="atmos",
        )
        with xc.set_options(metadata_locales="fr"):
            out = xs.compute_indicators(ds, [("tg_mean", indicator)])["YS-JAN"]
        out = xs.climatological_op(out, op="mean")

        assert (
            out["tg_mean_clim_mean"].attrs["long_name"]
            == "30-year climatological average of Mean daily mean temperature."
        )
        assert (
            out["tg_mean_clim_mean"].attrs["long_name_fr"]
            == "Moyenne 30 ans de Moyenne de la température moyenne quotidienne."
        )

    @pytest.mark.parametrize("locale", ["fr", "jp"])
    def test_add(self, locale):
        # Dummy function to make gettext aware of translatable-strings
        def _(s):
            return s

        ds = timeseries(
            np.arange(1, 366),
            variable="tas",
            start="2001-01-01",
            freq="D",
            as_dataset=True,
        )
        with xc.set_options(metadata_locales=locale):
            xs.utils.add_attr(ds, "some_attr", _("Ranking of measure performance"))
        assert ds.attrs["some_attr"] == "Ranking of measure performance"
        if locale == "fr":
            assert ds.attrs["some_attr_fr"] == "Classement de performance de la mesure"
        elif locale == "jp":
            # Japanese translation is not available, so the original string is used
            assert ds.attrs["some_attr_jp"] == "Ranking of measure performance"


class TestDateParser:
    @pytest.mark.parametrize(
        "date,end_of_period,dtype,exp",
        [
            ("2001", True, "datetime", pd.Timestamp("2001-12-31 23:59:59")),
            ("150004", True, "datetime", pd.Timestamp("1500-04-30 23:59:59")),
            ("31231212", None, "datetime", pd.Timestamp("3123-12-12")),
            ("2001-07-08", None, "period", pd.Period("2001-07-08", "h")),
            (pd.Timestamp("1993-05-20T12:07"), None, "str", "1993-05-20"),
            (
                cftime.Datetime360Day(1981, 2, 30),
                None,
                "datetime",
                pd.Timestamp("1981-02-28"),
            ),
            (
                np.datetime64("1200-11-12"),
                "Y",
                "datetime",
                pd.Timestamp("1200-12-31 23:59:59"),
            ),
            (
                datetime(2045, 5, 2, 13, 45),
                None,
                "datetime",
                pd.Timestamp("2045-05-02T13:45"),
            ),
            ("abc", None, "datetime", pd.Timestamp("NaT")),
            ("", True, "datetime", pd.Timestamp("NaT")),
            (
                pd.Period("2001-07-08", "h"),
                None,
                "datetime",
                pd.Timestamp("2001-07-08"),
            ),
        ],
    )
    def test_normal(self, date, end_of_period, dtype, exp):
        out = xs.utils.date_parser(date, end_of_period=end_of_period, out_dtype=dtype)
        if pd.isna(exp):
            assert pd.isna(out)
        else:
            assert out == exp


class TestMinCal:
    @pytest.mark.parametrize(
        "cals",
        [
            ["360_day", "365_day"],
            ["365_day", "default"],
            ["noleap", "default"],
            ["365_day", "noleap"],
            ["365_day", "all_leap"],
            ["366_day", "all_leap"],
            ["366_day", "default"],
        ],
    )
    def test_minimum_calendar(self, cals):
        out = xs.utils.minimum_calendar(cals)
        if "360_day" in cals:
            assert out == "360_day"
        elif any(c in cals for c in ["noleap", "365_day"]):
            assert out == "noleap"
        elif any(c in cals for c in ["default", "standard"]):
            assert out == "standard"
        else:
            assert out == "all_leap"

    def test_error(self):
        with pytest.warns(UserWarning, match="These calendars are not recognized"):
            xs.utils.minimum_calendar(["366_day", "foo"])


class TestTranslateTimeChunk:
    @pytest.mark.parametrize("chunk", [-1, 10])
    def test_normal(self, chunk):
        out = xs.utils.translate_time_chunk({"time": chunk, "lon": 50}, "noleap", 3450)
        assert out == {"time": 3450 if chunk == -1 else 10, "lon": 50}

    @pytest.mark.parametrize("calendar", ["360_day", "standard", "365_day", "366_day"])
    def test_ny(self, calendar):
        ndays = int(calendar.split("_")[0]) if "day" in calendar else 365.25
        out = xs.utils.translate_time_chunk(
            {"time": "4year", "lon": 50}, calendar, 3450
        )
        assert out == {"time": ndays * 4, "lon": 50}

    def test_warning(self):
        with pytest.warns(UserWarning, match="The number of days"):
            xs.utils.translate_time_chunk(
                {"time": "3year", "lon": 50}, "standard", 3450
            )

    def test_dict_of_dict(self):
        out = xs.utils.translate_time_chunk(
            {"tas": {"time": 10, "lon": 50}, "pr": {"time": -1, "lon": 50}},
            "noleap",
            3450,
        )
        assert out == {"tas": {"time": 10, "lon": 50}, "pr": {"time": 3450, "lon": 50}}


def test_naturalsort():
    assert xs.utils.natural_sort(["r1i1p1", "r2i1p1", "r10i1p1", "r1i1p2"]) == [
        "r1i1p1",
        "r1i1p2",
        "r2i1p1",
        "r10i1p1",
    ]


def get_cat_attrs():
    ds = timeseries(
        np.tile(np.arange(1, 2), 50),
        variable="tas",
        start="2000-01-01",
        freq="YS-JAN",
        as_dataset=True,
    )
    ds.attrs = {
        "foo": "bar",
        "cat:type": "simulation",
        "cat:variable": ("tas",),
        "dog:source": "CanESM5",
    }

    assert xs.utils.get_cat_attrs(ds) == {"type": "simulation", "variable": ("tas",)}
    assert xs.utils.get_cat_attrs(ds, var_as_str=True) == {
        "type": "simulation",
        "variable": "tas",
    }
    assert xs.utils.get_cat_attrs(ds, prefix="dog:") == {"source": "CanESM5"}
    assert xs.utils.get_cat_attrs(ds.attrs) == {
        "type": "simulation",
        "variable": ("tas",),
    }


class TestScripting:
    ds = timeseries(
        np.tile(np.arange(1, 2), 50),
        variable="tas",
        start="2000-01-01",
        freq="YS-JAN",
        as_dataset=True,
    )
    ds.attrs = {
        "cat:type": "simulation",
        "cat:processing_level": "raw",
        "cat:variable": ("tas",),
        "dog:source": "CanESM5",
    }

    @pytest.mark.parametrize(
        "ds, prefix, var_as_str",
        [["ds", "cat:", False], ["dict", "cat:", True], ["ds", "dog:", True]],
    )
    def test_get_cat_attrs(self, ds, prefix, var_as_str):
        data = self.ds if ds == "ds" else self.ds.attrs
        out = xs.utils.get_cat_attrs(data, prefix=prefix, var_as_str=var_as_str)

        if var_as_str and prefix == "cat:":
            assert out == {
                "type": "simulation",
                "processing_level": "raw",
                "variable": "tas",
            }
        elif not var_as_str and prefix == "cat:":
            assert out == {
                "type": "simulation",
                "processing_level": "raw",
                "variable": ("tas",),
            }
        elif prefix == "dog:":
            assert out == {"source": "CanESM5"}

    def test_strip_cat_attrs(self):
        out = xs.utils.strip_cat_attrs(self.ds)
        assert list(out.attrs.keys()) == ["dog:source"]


class TestStack:
    def test_no_nan(self):
        ds = datablock_3d(
            np.zeros((20, 10, 10)),
            "tas",
            "lon",
            -5,
            "lat",
            80.5,
            1,
            1,
            "2000-01-01",
            as_dataset=True,
        )
        mask = xr.where(ds.tas.isel(time=0).isnull(), False, True).drop_vars("time")
        out = xs.utils.stack_drop_nans(ds, mask=mask)
        assert "loc" in out.dims
        assert out.sizes["loc"] == 100

        ds_unstack = xs.utils.unstack_fill_nan(out)
        assert ds_unstack.equals(ds)

    def test_nan(self, tmp_path):
        data = np.zeros((20, 10, 10))
        data[:, 0, 0] = [np.nan] * 20
        ds = datablock_3d(
            data,
            "tas",
            "lon",
            -5,
            "lat",
            80.5,
            1,
            1,
            "2000-01-01",
            as_dataset=True,
        )

        mask = xr.where(ds.tas.isel(time=0).isnull(), False, True).drop_vars("time")
        ds.attrs["cat:domain"] = "RegionEssai"
        out = xs.utils.stack_drop_nans(
            ds,
            mask=mask,
            new_dim="loc1",
            to_file=str(tmp_path / "subfolder" / "coords_{domain}_{shape}.nc"),
        )
        assert "loc1" in out.dims
        assert out.sizes["loc1"] == 99
        assert (tmp_path / "subfolder" / "coords_RegionEssai_10x10.nc").is_file()

        out_no_mask = xs.utils.stack_drop_nans(ds, mask=["lon", "lat"], new_dim="loc1")
        assert out_no_mask.equals(out)

        ds_unstack = xs.utils.unstack_fill_nan(
            out,
            dim="loc1",
            coords=str(tmp_path / "subfolder" / "coords_{domain}_{shape}.nc"),
        )
        assert ds_unstack.equals(ds)

    @pytest.mark.parametrize("coords", ["file.nc", ["lon", "lat"], "dict", None])
    def test_fillnan_coords(self, tmpdir, coords):
        data = np.zeros((20, 10, 10))
        data[:, 1, 0] = [np.nan] * 20
        data[:, 0, :] = [np.nan] * 10
        ds = datablock_3d(
            data,
            "tas",
            "lon",
            -5,
            "lat",
            80.5,
            1,
            1,
            "2000-01-01",
            as_dataset=True,
        )
        ds.attrs["cat:domain"] = "RegionEssai"

        mask = xr.where(ds.tas.isel(time=0).isnull(), False, True).drop_vars("time")
        # Add mask as a coordinate
        ds = ds.assign_coords(z=mask.astype(int))
        ds.z.attrs["foo"] = "bar"

        if coords == "dict":
            coords = {"lon": ds.lon, "lat": ds.lat, "z": ds.z}
        elif coords == "file.nc":
            coords = str(tmpdir / "coords_{domain}_{shape}.nc")

        ds_stack = xs.utils.stack_drop_nans(
            ds, mask=mask, to_file=coords if isinstance(coords, str) else None
        )
        ds_unstack = xs.utils.unstack_fill_nan(
            ds_stack,
            coords=coords,
        )

        if isinstance(coords, list):
            # Cannot fully recover the original dataset.
            ds_unstack["z"] = ds_unstack["z"].fillna(0)
            assert ds_unstack.equals(ds.isel(lat=slice(1, None)))
        elif coords is None:
            # 'z' gets completely assigned as a dimension.
            assert "z" in ds_unstack.dims
            assert (
                ds_unstack.isel(z=0)
                .drop_vars("z")
                .equals(ds.isel(lat=slice(1, None)).drop_vars("z"))
            )
        else:
            assert ds_unstack.equals(ds)

    def test_maybe(self, tmp_path):
        data = np.zeros((20, 10, 10))
        data[:, 0, 0] = [np.nan] * 20
        ds = datablock_3d(
            data,
            "tas",
            "lon",
            -5,
            "lat",
            80.5,
            1,
            1,
            "2000-01-01",
            as_dataset=True,
        )
        mask = xr.where(ds.tas.isel(time=0).isnull(), False, True).drop_vars("time")
        ds.attrs["cat:domain"] = "RegionEssai"
        z = xr.DataArray(
            np.ones([10, 10]),
            dims=["lat", "lon"],
            coords={"lat": ds.lat, "lon": ds.lon},
        )
        z1d = xr.DataArray(np.ones([10]), dims=["lat"], coords={"lat": ds.lat})
        ds = ds.assign_coords(z=z, z1d=z1d)
        out = xs.utils.stack_drop_nans(
            ds,
            mask=mask,
            new_dim="loc1",
            to_file=str(tmp_path / "coords_{domain}_{shape}.nc"),
        )

        maybe_unstacked = xs.utils.maybe_unstack(
            out, dim="loc1", coords=str(tmp_path / "coords_{domain}_{shape}.nc")
        )
        assert maybe_unstacked.equals(out)
        # Call through clean_up to test the whole pipeline
        maybe_unstack_dict = {
            "dim": "loc1",
            "coords": str(tmp_path / "coords_{domain}_{shape}.nc"),
            "stack_drop_nans": True,
        }
        maybe_unstacked = xs.utils.clean_up(out, maybe_unstack_dict=maybe_unstack_dict)
        assert maybe_unstacked.equals(ds)
        maybe_unstacked = xs.utils.maybe_unstack(
            out,
            dim="loc1",
            coords=str(tmp_path / "coords_{domain}_{shape}.nc"),
            rechunk={"lon": -1, "lat": 2},
            stack_drop_nans=True,
        )
        assert dict(maybe_unstacked.chunks) == {
            "time": (20,),
            "lat": (2, 2, 2, 2, 2),
            "lon": (10,),
        }


class TestVariablesUnits:
    def test_variables_same(self):
        ds = timeseries(
            np.tile(np.arange(1, 13), 3),
            variable="tas",
            start="2001-01-01",
            freq="MS",
            as_dataset=True,
        )
        out = xs.clean_up(ds, variables_and_units={"tas": "degK"})
        assert out.tas.attrs["units"] == "degK"
        np.testing.assert_array_equal(out.tas, ds.tas)

    def test_variables_2(self):
        ds = timeseries(
            np.tile(np.arange(1, 13), 3),
            variable="tas",
            start="2001-01-01",
            freq="MS",
            as_dataset=True,
        )
        ds["pr"] = timeseries(
            np.tile(np.arange(1, 13), 3),
            variable="pr",
            start="2001-01-01",
            freq="MS",
        )
        out = xs.clean_up(ds, variables_and_units={"tas": "degK"})
        assert out.tas.attrs["units"] == "degK"
        assert out.pr.attrs["units"] == "kg m-2 s-1"

    def test_variables_sametime(self):
        ds = timeseries(
            np.tile(np.arange(1, 13), 3),
            variable="pr",
            start="2001-01-01",
            freq="D",
            as_dataset=True,
        )
        out = xs.clean_up(ds, variables_and_units={"pr": "mm/day"})
        assert out.pr.attrs["units"] == "mm d-1"
        np.testing.assert_array_almost_equal(out.pr, ds.pr * 86400)

    def test_variables_amount2rate(self):
        ds = timeseries(
            np.tile(np.arange(1, 13), 3),
            variable="pr",
            start="2001-01-01",
            freq="D",
            units="mm",
            as_dataset=True,
        )
        out = xs.clean_up(ds, variables_and_units={"pr": "mm/day"})
        assert out.pr.attrs["units"] == "mm d-1"
        np.testing.assert_array_almost_equal(out.pr, ds.pr)

    def test_variables_rate2amount(self):
        ds = timeseries(
            np.tile(np.arange(1, 13), 3),
            variable="pr",
            start="2001-01-01",
            freq="D",
            as_dataset=True,
        )
        out = xs.clean_up(ds, variables_and_units={"pr": "mm"})
        assert out.pr.attrs["units"] == "mm"
        np.testing.assert_array_almost_equal(out.pr, ds.pr * 86400)

    def test_variables_error(self):
        ds = timeseries(
            np.tile(np.arange(1, 13), 3),
            variable="pr",
            start="2001-01-01",
            freq="D",
            units="mm s-2",
            as_dataset=True,
        )
        with pytest.raises(ValueError, match="No known transformation"):
            xs.clean_up(ds, variables_and_units={"pr": "mm"})


class TestCalendar:
    def test_normal(self):
        ds = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="tas",
            start="2000-01-01",
            freq="D",
            as_dataset=True,
        )

        out = xs.clean_up(ds, convert_calendar_kwargs={"calendar": "noleap"})
        assert isinstance(out.time.values[0], cftime.DatetimeNoLeap)
        assert len(out.time) == 365 * 4

    def test_360(self):
        ds = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="tas",
            start="2000-01-01",
            freq="D",
            as_dataset=True,
        )

        out = xs.clean_up(ds, convert_calendar_kwargs={"calendar": "360_day"})
        assert isinstance(out.time.values[0], cftime.Datetime360Day)
        assert len(out.time) == 360 * 4
        assert len(out.time.sel(time="2000-02-30")) == 1

    def test_missing_by_var(self):
        ds = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="tas",
            start="2000-01-01",
            freq="D",
            as_dataset=True,
        )
        ds["pr"] = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="pr",
            start="2000-01-01",
            freq="D",
        )
        ds = xs.clean_up(ds, convert_calendar_kwargs={"calendar": "noleap"})
        missing_by_vars = {"tas": "interpolate", "pr": 9999}

        out = xs.clean_up(
            ds,
            convert_calendar_kwargs={"calendar": "standard"},
            missing_by_var=missing_by_vars,
        )
        assert out.tas.isnull().sum() == 0
        np.testing.assert_array_equal(out.tas.sel(time="2000-02-29"), 60)
        assert out.pr.isnull().sum() == 0
        np.testing.assert_array_equal(out.pr.sel(time="2000-02-29"), 9999)

    def test_missing_by_var_error(self):
        ds = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="tas",
            start="2000-01-01",
            freq="D",
            as_dataset=True,
        )
        ds["pr"] = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="pr",
            start="2000-01-01",
            freq="D",
        )
        missing_by_vars = {"pr": 9999}
        with pytest.raises(ValueError, match="All variables must be"):
            xs.clean_up(
                ds,
                convert_calendar_kwargs={"calendar": "standard"},
                missing_by_var=missing_by_vars,
            )


def test_round():
    ds = timeseries(
        np.arange(1, 365 * 4 + 2) / 1234,
        variable="tas",
        start="2000-01-01",
        freq="D",
        as_dataset=True,
    )
    ds["pr"] = timeseries(
        np.arange(1, 365 * 4 + 2) / 1234,
        variable="pr",
        start="2000-01-01",
        freq="D",
    )
    out = xs.clean_up(ds, round_var={"tas": 6, "pr": 1})
    np.testing.assert_array_equal(out.tas.isel(time=0), 0.000810)
    np.testing.assert_array_equal(out.pr.isel(time=0), 0.0)
    assert "Rounded 'pr' to 1 decimal" in out["pr"].attrs["history"]


class TestAttrs:
    def test_common(self):
        ds1 = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="tas",
            start="2000-01-01",
            freq="D",
            as_dataset=True,
        )

        ds2 = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="tas",
            start="2000-01-01",
            freq="D",
            as_dataset=True,
        )
        ds2.attrs = {
            "foo": "bar",
            "cat:type": "simulation",
            "cat:variable": ("tas",),
            "cat:source": "CNRM-CM6",
            "cat:mip_era": "CMIP6",
        }
        ds3 = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="tas",
            start="2000-01-01",
            freq="D",
            as_dataset=True,
        )
        ds3.attrs = {
            "foo": "bar",
            "cat:type": "simulation",
            "cat:variable": ("tas",),
            "cat:source": "CanESM5",
            "cat:mip_era": "CMIP6",
        }

        # Nothing in common between ds1 and the other datasets
        ds1.attrs = {"bar": "foo"}
        out = xs.clean_up(ds1, common_attrs_only=[ds2, ds3])
        assert out.attrs == {}

        ds1.attrs = ds2.attrs
        out = xs.clean_up(ds1, common_attrs_only=[ds2, ds3])
        assert all(
            k in out.attrs
            for k in ["foo", "cat:type", "cat:variable", "cat:mip_era", "cat:id"]
        )
        assert out.attrs["cat:id"] == "CMIP6"

        del ds1.attrs["cat:mip_era"]
        out = xs.clean_up(ds1, common_attrs_only={"a": ds2, "b": ds3})
        assert all(
            k in out.attrs for k in ["foo", "cat:type", "cat:variable", "cat:id"]
        )
        assert out.attrs["cat:id"] == ""

    @pytest.mark.requires_netcdf
    def test_common_open(self):
        ds1 = xr.open_dataset(
            Path(__file__).parent.parent
            / "docs"
            / "notebooks"
            / "samples"
            / "tutorial"
            / "ScenarioMIP"
            / "example-region"
            / "NCC"
            / "NorESM2-MM"
            / "ssp126"
            / "r1i1p1f1"
            / "day"
            / "ScenarioMIP_NCC_NorESM2-MM_ssp126_r1i1p1f1_gn_raw.nc"
        )
        ds1.attrs["cat:id"] = "SomeID"
        ds2 = (
            Path(__file__).parent.parent
            / "docs"
            / "notebooks"
            / "samples"
            / "tutorial"
            / "ScenarioMIP"
            / "example-region"
            / "NCC"
            / "NorESM2-MM"
            / "ssp245"
            / "r1i1p1f1"
            / "day"
            / "ScenarioMIP_NCC_NorESM2-MM_ssp245_r1i1p1f1_gn_raw.nc"
        )
        ds3 = (
            Path(__file__).parent.parent
            / "docs"
            / "notebooks"
            / "samples"
            / "tutorial"
            / "ScenarioMIP"
            / "example-region"
            / "NCC"
            / "NorESM2-MM"
            / "ssp585"
            / "r1i1p1f1"
            / "day"
            / "ScenarioMIP_NCC_NorESM2-MM_ssp585_r1i1p1f1_gn_raw.nc"
        )

        out = xs.clean_up(ds1, common_attrs_only=[ds2, ds3])
        assert (
            out.attrs["comment"]
            == "This is a test file created for the xscen tutorial. This file is not a real CMIP6 file."
        )
        assert out.attrs.get("cat:id") is None

    def test_to_level(self):
        ds = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="tas",
            start="2000-01-01",
            freq="D",
            as_dataset=True,
        )
        out = xs.clean_up(ds, to_level="cat")
        assert out.attrs["cat:processing_level"] == "cat"

    def test_remove(self):
        ds = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="tas",
            start="2000-01-01",
            freq="D",
            as_dataset=True,
        )
        ds.attrs = {
            "foo": "bar",
            "cat:type": "simulation",
            "cat:variable": ("tas",),
            "cat:source": "CNRM-CM6",
            "bacat:mip_era": "CMIP6",
        }
        out = xs.clean_up(
            ds, attrs_to_remove={"tas": ["units"], "global": [".*cat.*", "foo"]}
        )
        assert "units" in ds.tas.attrs
        assert "units" not in out.tas.attrs
        assert out.attrs == {}
        out2 = xs.clean_up(ds, attrs_to_remove={"tas": ["units"], "global": ["cat.*"]})
        assert out2.attrs == {"foo": "bar", "bacat:mip_era": "CMIP6"}

    def test_remove_except(self):
        ds = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="tas",
            start="2000-01-01",
            freq="D",
            as_dataset=True,
        )
        ds.attrs = {
            "foo": "bar",
            "cat:type": "simulation",
            "cat:variable": ("tas",),
            "cat:source": "CNRM-CM6",
            "bacat:mip_era": "CMIP6",
        }
        out = xs.clean_up(
            ds, remove_all_attrs_except={"tas": ["units"], "global": [".*cat.*"]}
        )
        assert out.tas.attrs == {"units": "K"}
        assert out.attrs == {
            "cat:type": "simulation",
            "cat:variable": ("tas",),
            "cat:source": "CNRM-CM6",
            "bacat:mip_era": "CMIP6",
        }
        out2 = xs.clean_up(ds, remove_all_attrs_except={"global": ["cat.*"]})
        assert len(out2.tas.attrs) == 4
        assert out2.attrs == {
            "cat:type": "simulation",
            "cat:variable": ("tas",),
            "cat:source": "CNRM-CM6",
        }

    def test_add_attr(self):
        ds = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="tas",
            start="2000-01-01",
            freq="D",
            as_dataset=True,
        )
        out = xs.clean_up(
            ds,
            add_attrs={"tas": {"foo": "bar"}, "global": {"foo2": "electric boogaloo"}},
        )
        assert out.tas.attrs["foo"] == "bar"
        assert out.attrs["foo2"] == "electric boogaloo"

    @pytest.mark.parametrize(
        "change_prefix", ["dog", {"cat": "dog:"}, {"cat:": "dog:", "bacat": "badog"}]
    )
    def test_change_prefix(self, change_prefix):
        ds = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="tas",
            start="2000-01-01",
            freq="D",
            as_dataset=True,
        )
        ds.attrs = {
            "foo": "bar",
            "cat:type": "simulation",
            "cat:variable": ("tas",),
            "cat:source": "CNRM-CM6",
            "bacat:mip_era": "CMIP6",
        }
        out = xs.clean_up(ds, change_attr_prefix=change_prefix)
        if isinstance(change_prefix, str) or len(change_prefix) == 1:
            assert out.attrs == {
                "foo": "bar",
                "dog:type": "simulation",
                "dog:variable": ("tas",),
                "dog:source": "CNRM-CM6",
                "bacat:mip_era": "CMIP6",
            }
        else:
            assert out.attrs == {
                "foo": "bar",
                "dog:type": "simulation",
                "dog:variable": ("tas",),
                "dog:source": "CNRM-CM6",
                "badog:mip_era": "CMIP6",
            }


class TestUnstackDates:
    @pytest.mark.parametrize(
        "freq", ["MS", "2MS", "3MS", "QS-DEC", "QS", "2QS", "YS", "YS-DEC", "4YS"]
    )
    def test_normal(self, freq):
        ds = timeseries(
            np.arange(1, 35),
            variable="tas",
            start="2000-01-01",
            freq=freq,
            as_dataset=True,
        )
        out = xs.utils.unstack_dates(ds)
        np.testing.assert_array_equal(
            out.time,
            pd.date_range(
                "2000-01-01",
                periods=len(np.unique(ds.time.dt.year)),
                freq="YS" if freq != "4YS" else "4YS",
            ),
        )
        if freq == "MS":
            np.testing.assert_array_equal(
                out.month,
                [
                    "JAN",
                    "FEB",
                    "MAR",
                    "APR",
                    "MAY",
                    "JUN",
                    "JUL",
                    "AUG",
                    "SEP",
                    "OCT",
                    "NOV",
                    "DEC",
                ],
            )
        elif "M" in freq:
            assert len(out.season) == 12 / int(freq[0])
            np.testing.assert_array_equal(
                out.season[0], ["JF"] if freq == "2MS" else ["JFM"]
            )
        elif "QS" in freq:
            assert len(out.season) == 4 if freq != "2QS" else 2
            np.testing.assert_array_equal(
                out.season[0],
                (
                    ["MAM"]
                    if freq == "QS-DEC"
                    else ["JFM"] if freq == "QS" else ["JFMAMJ"]
                ),
            )
        else:
            assert len(out.season) == 1
            np.testing.assert_array_equal(
                out.season[0], [f"{freq.replace('YS', 'annual')}"]
            )

    @pytest.mark.parametrize("freq", ["MS", "QS", "YS"])
    def test_seasons(self, freq):
        ds = timeseries(
            np.arange(1, 35),
            variable="tas",
            start="2000-01-01",
            freq=freq,
            as_dataset=True,
        )
        seasons = {
            1: "january",
            2: "february",
            3: "march",
            4: "april",
            5: "may",
            6: "june",
            7: "july",
            8: "august",
            9: "september",
            10: "october",
            11: "november",
            12: "december",
        }
        out = xs.utils.unstack_dates(ds, seasons=seasons)
        if freq == "MS":
            np.testing.assert_array_equal(
                out.month,
                [
                    "january",
                    "february",
                    "march",
                    "april",
                    "may",
                    "june",
                    "july",
                    "august",
                    "september",
                    "october",
                    "november",
                    "december",
                ],
            )
        elif freq == "QS":
            np.testing.assert_array_equal(
                out.season, ["january", "april", "july", "october"]
            )
        elif freq == "YS":
            np.testing.assert_array_equal(out.season, ["january"])

    @pytest.mark.parametrize("freq", ["2MS", "QS-DEC", "YS-DEC"])
    def test_winter(self, freq):
        ds = timeseries(
            np.arange(1, 35),
            variable="tas",
            start="2000-12-01",
            freq=freq,
            as_dataset=True,
        )
        out = xs.utils.unstack_dates(ds)
        assert pd.Timestamp(str(out.time[0].values).split("T")[0]) == pd.Timestamp(
            "2000-01-01"
        )
        out = xs.utils.unstack_dates(ds, winter_starts_year=True)
        assert pd.Timestamp(str(out.time[0].values).split("T")[0]) == pd.Timestamp(
            "2001-01-01"
        )

    def test_coords(self):
        freq = "MS"
        ds = timeseries(
            np.arange(1, 35),
            variable="tas",
            start="2000-01-01",
            freq=freq,
        )
        ds["horizon"] = xr.DataArray(
            np.array(["2001-2009"] * len(ds.time)),
            dims="time",
            coords={"time": ds.time},
        )
        ds = ds.assign_coords({"horizon": ds.horizon})
        out = xs.utils.unstack_dates(ds)
        assert all(k in out["horizon"].dims for k in ["time", "month"])

    def test_dask(self):
        freq = "MS"
        ds = timeseries(
            dask.array.from_array(np.arange(1, 35), chunks=(10,)),
            variable="tas",
            start="2000-01-01",
            freq=freq,
        )
        out = xs.utils.unstack_dates(ds)
        assert isinstance(out.data, dask.array.Array)
        assert "month" in out.dims

    def test_errors(self):
        ds = timeseries(
            np.arange(1, 365 * 4 + 2),
            variable="tas",
            start="2000-01-01",
            freq="D",
            as_dataset=True,
        )
        with pytest.raises(ValueError, match="Only monthly frequencies"):
            xs.utils.unstack_dates(ds)
        ds = ds.where(ds.time.dt.day != 1, drop=True)
        with pytest.raises(
            ValueError, match="The data must have a clean time coordinate."
        ):
            xs.utils.unstack_dates(ds)

        ds = timeseries(
            np.arange(1, 13),
            variable="tas",
            start="2000-01-01",
            freq="7MS",
            as_dataset=True,
        )
        with pytest.raises(
            ValueError, match="Only periods that divide the year evenly are supported."
        ):
            xs.utils.unstack_dates(ds)


class TestEnsureTime:
    def test_xrfreq_ok(self):
        ds = timeseries(
            np.arange(1, 360),
            variable="tas",
            start="2000-01-01T12:00:00",
            freq="D",
            as_dataset=True,
        )
        out = xs.utils.ensure_correct_time(ds, "D")
        assert np.all(out.time.dt.hour == 0)

    def test_xrfreq_bad(self):
        ds = timeseries(
            np.arange(1, 360),
            variable="tas",
            start="2000-01-01T12:00:00",
            freq="D",
            as_dataset=True,
        )
        # Add random small number of seconds to the time
        ds["time"] = ds.time + (np.random.rand(len(ds.time)) * 10).astype(
            "timedelta64[s]"
        )
        out = xs.utils.ensure_correct_time(ds, "D")
        assert np.all(out.time.dt.hour == 0)
        assert np.all(out.time.dt.second == 0)

    def test_xrfreq_error(self):
        ds = timeseries(
            np.arange(1, 360),
            variable="tas",
            start="2000-01-01T12:00:00",
            freq="D",
            as_dataset=True,
        )
        # Add random small number of seconds to the time
        rng = np.random.default_rng(0)
        ds["time"] = ds.time + (rng.random(len(ds.time)) * 24).astype("timedelta64[h]")
        with pytest.raises(
            ValueError,
            match="Dataset is labelled as having a sampling frequency of D, but some periods have more than one data point.",
        ):
            xs.utils.ensure_correct_time(ds, "D")
        ds = timeseries(
            np.arange(1, 360),
            variable="tas",
            start="2000-01-01T12:00:00",
            freq="D",
            as_dataset=True,
        )
        # Remove some time points
        ds = ds.where(ds.time.dt.day % 2 == 0, drop=True)
        with pytest.raises(
            ValueError,
            match="The resampling count contains NaNs or 0s. There might be some missing data.",
        ):
            xs.utils.ensure_correct_time(ds, "D")


class TestStandardPeriod:
    @pytest.mark.parametrize(
        "period, timestamp_eop_match, timestamp_neop_match",
        [
            ([1981, 2000], "2000-12-31 23:59:59", "2000-01-01"),
            ([[1981, 2000]], "2000-12-31 23:59:59", "2000-01-01"),
            (["1981", "2000"], "2000-12-31 23:59:59", "2000-01-01"),
            ([["1981", "2000-02"]], "2000-02-29 23:59:59", "2000-02-01"),
            ([["1981", "2000-12"]], "2000-12-31 23:59:59", "2000-12-01"),
            (
                [[pd.Timestamp("1981"), pd.Timestamp("2000")]],
                "2000-01-01 00:00:00",
                "2000-01-01",
            ),
        ],
    )
    def test_normal(self, period, timestamp_eop_match, timestamp_neop_match):
        out = xs.utils.standardize_periods(period, multiple=True)
        assert out == [["1981", "2000"]]

        out = xs.utils.standardize_periods(period, multiple=False)
        assert out == ["1981", "2000"]

        out = xs.utils.standardize_periods(period, multiple=False, out_dtype="datetime")
        assert out == [pd.Timestamp("1981-01-01"), pd.Timestamp(timestamp_eop_match)]

        out = xs.utils.standardize_periods(
            period, multiple=False, out_dtype="datetime", end_of_periods=False
        )
        assert out == [pd.Timestamp("1981-01-01"), pd.Timestamp(timestamp_neop_match)]

    def test_error(self):
        assert xs.utils.standardize_periods(None) is None
        with pytest.raises(ValueError, match="should be comprised of two elements"):
            xs.utils.standardize_periods(["1981-2010"])
        with pytest.raises(ValueError, match="should be in chronological order,"):
            xs.utils.standardize_periods(["2010", "1981"])
        with pytest.raises(ValueError, match="should be a single instance"):
            xs.utils.standardize_periods(
                [["1981", "2010"], ["1981", "2010"]], multiple=False
            )


def test_sort_seasons():
    seasons = pd.Index(["JJA", "DJF", "SON", "MAM"])
    out = xs.utils.season_sort_key(seasons, name="season")
    np.testing.assert_array_equal(out, [6, 0, 9, 3])

    seasons = pd.Index(["JFM", "JAS", "OND", "AMJ"])
    out = xs.utils.season_sort_key(seasons, name="season")
    np.testing.assert_array_equal(out, [1, 7, 10, 4])

    seasons = pd.Index(["FEB", "JAN", "MAR", "DEC"])
    out = xs.utils.season_sort_key(seasons, name="month")
    np.testing.assert_array_equal(out, [1, 0, 2, 11])

    # Invalid returns the original object
    seasons = pd.Index(["FEB", "DEC", "foo"])
    out = xs.utils.season_sort_key(seasons, name="month")
    np.testing.assert_array_equal(out, seasons)


def test_xrfreq_to_timedelta():
    assert xs.utils.xrfreq_to_timedelta("D") == pd.Timedelta(1, "D")
    assert xs.utils.xrfreq_to_timedelta("QS-DEC") == pd.Timedelta(90, "D")
    assert xs.utils.xrfreq_to_timedelta("YS") == pd.Timedelta(365, "D")
    assert xs.utils.xrfreq_to_timedelta("2QS") == pd.Timedelta(180, "D")
    with pytest.raises(ValueError, match="Invalid frequency"):
        xs.utils.xrfreq_to_timedelta("foo")


def test_ensure_new_xrfreq():
    assert xs.utils.ensure_new_xrfreq("2M") == "2ME"
    assert xs.utils.ensure_new_xrfreq("2Q-DEC") == "2QE-DEC"
    assert xs.utils.ensure_new_xrfreq("AS-JUL") == "YS-JUL"
    assert xs.utils.ensure_new_xrfreq("A-JUL") == "YE-JUL"
    assert xs.utils.ensure_new_xrfreq("Y-JUL") == "YE-JUL"
    assert xs.utils.ensure_new_xrfreq("A") == "YE"
    assert xs.utils.ensure_new_xrfreq("Y") == "YE"
    assert xs.utils.ensure_new_xrfreq("3H") == "3h"
    assert xs.utils.ensure_new_xrfreq("3T") == "3min"
    assert xs.utils.ensure_new_xrfreq("3S") == "3s"
    assert xs.utils.ensure_new_xrfreq("3L") == "3ms"
    assert xs.utils.ensure_new_xrfreq("3U") == "3us"

    # Errors
    assert xs.utils.ensure_new_xrfreq(3) == 3
    assert xs.utils.ensure_new_xrfreq("foo") == "foo"


def test_xarray_defaults():
    kwargs = {
        "chunks": {"time": 10},
        "foo": "bar",
        "xr_open_kwargs": {"decode_times": False},
        "xr_combine_kwargs": {"combine_attrs": "drop"},
    }
    out = xs.utils._xarray_defaults(**kwargs)
    assert out == {
        "chunks": {"time": 10},
        "foo": "bar",
        "xarray_open_kwargs": {"decode_times": False, "chunks": {}},
        "xarray_combine_by_coords_kwargs": {
            "combine_attrs": "drop",
            "data_vars": "minimal",
        },
    }



import contextlib
import warnings

import numpy as np
import pytest
import xarray as xr
import xclim as xc
import xsdba
from xclim.testing.helpers import test_timeseries as timeseries

# Copied from xarray/core/nputils.py
# Can be removed once numpy 2.0+ is the oldest supported version
try:
    from numpy.exceptions import RankWarning
except ImportError:
    from numpy import RankWarning

import xscen as xs


class TestTrain:
    dref = timeseries(
        np.ones(365 * 3), variable="tas", start="2001-01-01", freq="D", as_dataset=True
    )

    dhist = timeseries(
        np.concatenate([np.ones(365 * 2) * 2, np.ones(365) * 3]),
        variable="tas",
        start="2001-01-01",
        freq="D",
        as_dataset=True,
    )
    dhist.attrs["cat:xrfreq"] = "D"
    dhist.attrs["cat:domain"] = "one_point"
    dhist.attrs["cat:id"] = "fake_id"

    @pytest.mark.parametrize(
        "var, period",
        [("tas", ["2001", "2002"]), (["tas"], ["2003", "2003"])],
    )
    def test_basic_train(self, var, period):
        out = xs.train(self.dref, self.dhist, var=var, period=period)

        assert out.attrs["cat:xrfreq"] == "D"
        assert out.attrs["cat:domain"] == "one_point"
        assert out.attrs["cat:id"] == "fake_id"
        assert out.attrs["cat:processing_level"] == "training_tas"

        assert "dayofyear" in out
        assert "quantiles" in out
        result = [-1] * 365 if period == ["2001", "2002"] else [-2] * 365
        np.testing.assert_array_equal(out["scaling"], result)

    def test_preprocess(self):
        # FIXME: put back the test when xclim 0.55 is released, https://github.com/Ouranosinc/xclim/pull/2038/files
        # dhist360 = self.dhist.convert_calendar("360_day", align_on="year")
        dhist360 = self.dhist.convert_calendar("noleap", align_on="year")

        out = xs.train(
            self.dref,
            dhist360,
            var="tas",
            period=["2001", "2002"],
            adapt_freq={"thresh": "2 K"},
            jitter_over={"upper_bnd": "3 K", "thresh": "2 K"},
            jitter_under={"thresh": "2 K"},
        )

        assert out.attrs["train_params"] == {
            "maximal_calendar": "noleap",
            "adapt_freq": {"thresh": "2 K"},
            "jitter_over": {"upper_bnd": "3 K", "thresh": "2 K"},
            "jitter_under": {"thresh": "2 K"},
            "var": ["tas"],
        }

        assert "pth" in out
        assert "dP0" in out
        assert "dayofyear" in out
        assert "quantiles" in out

    def test_group(self):
        out = xs.train(
            self.dref,
            self.dhist,
            var="tas",
            period=["2001", "2002"],
            group={"group": "time.month", "window": 1},
        )

        out1 = xs.train(
            self.dref,
            self.dhist,
            var="tas",
            period=["2001", "2002"],
            group="time.month",
        )

        assert "month" in out
        assert "quantiles" in out
        assert out1.equals(out)

    def test_errors(self):
        with pytest.raises(ValueError):
            xs.train(self.dref, self.dhist, var=["tas", "pr"], period=["2001", "2002"])


class TestAdjust:
    dref = timeseries(
        np.ones((365 * 3) + 1),  # leap year
        variable="tas",
        start="2001-01-01",
        freq="D",
        as_dataset=True,
    )

    dsim = timeseries(
        np.concatenate([np.ones(365 * 3) * 2, np.ones((365 * 3) + 1) * 4]),
        variable="tas",
        start="2001-01-01",
        freq="D",
        as_dataset=True,
    )
    dsim.attrs["cat:xrfreq"] = "D"
    dsim.attrs["cat:domain"] = "one_point"
    dsim.attrs["cat:id"] = "fake_id"

    @pytest.mark.parametrize(
        "periods, to_level, bias_adjust_institution, bias_adjust_project",
        [
            (["2001", "2006"], None, None, None),
            ([["2001", "2001"], ["2006", "2006"]], "test", "i", "p"),
        ],
    )
    def test_basic(
        self, periods, to_level, bias_adjust_institution, bias_adjust_project
    ):
        dtrain = xs.train(
            self.dref.copy(),
            self.dsim.sel(time=slice("2001", "2003")),
            var="tas",
            period=["2001", "2003"],
        )

        # For justification of warning filter, see: https://docs.xarray.dev/en/stable/generated/xarray.Dataset.polyfit.html
        if (
            to_level is None
            and bias_adjust_institution is None
            and bias_adjust_project is None
        ):
            # No warning is expected
            context = contextlib.nullcontext()
        else:
            context = pytest.warns(RankWarning)
        with context:
            out = xs.adjust(
                dtrain,
                self.dsim.copy(),
                periods=periods,
                to_level=to_level,
                bias_adjust_institution=bias_adjust_institution,
                bias_adjust_project=bias_adjust_project,
            )

        assert out.attrs["cat:processing_level"] == to_level or "biasadjusted"
        assert out.attrs["cat:variable"] == ("tas",)
        assert out.attrs["cat:id"] == "fake_id"
        assert (
            out["tas"].attrs["bias_adjustment"]
            == "DetrendedQuantileMapping(group=Grouper("
            "name='time.dayofyear', window=31), kind='+'"
            ").adjust(sim, )"
        )
        assert out.time.dt.calendar == "noleap"

        if bias_adjust_institution is not None:
            assert out.attrs["cat:bias_adjust_institution"] == "i"
        if bias_adjust_project is not None:
            assert out.attrs["cat:bias_adjust_project"] == "p"

        assert out.time.dt.year.values[0] == 2001
        assert out.time.dt.year.values[-1] == 2006

        if periods == ["2001", "2006"]:
            np.testing.assert_array_equal(
                out["tas"].values,
                np.concatenate(
                    [np.ones(365 * 3) * 1, np.ones(365 * 3) * 3]
                ),  # -1 for leap year
            )
        else:  # periods==[['2001','2001'], ['2006','2006']]
            np.testing.assert_array_equal(
                out["tas"].values,
                np.concatenate([np.ones(365 * 1) * 1, np.ones(365 * 1) * 3]),
            )

    def test_write_train(self, tmpdir):
        dtrain = xs.train(
            self.dref.copy(),
            self.dsim.sel(time=slice("2001", "2003")),
            var="tas",
            period=["2001", "2003"],
            adapt_freq={"thresh": "2 K"},
            jitter_over={"upper_bnd": "3 K", "thresh": "2 K"},
            jitter_under={"thresh": "2 K"},
        )

        root = str(tmpdir / "_data")
        xs.save_to_zarr(dtrain, f"{root}/test.zarr", mode="o")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            dtrain2 = xr.open_dataset(
                f"{root}/test.zarr", chunks={"dayofyear": 365, "quantiles": 15}
            )

        out = xs.adjust(
            dtrain,
            self.dsim.copy(),
            periods=["2001", "2006"],
            xsdba_adjust_args={
                "detrend": {
                    "LoessDetrend": {"f": 0.2, "niter": 1, "d": 0, "weights": "tricube"}
                }
            },
        )

        out2 = xs.adjust(
            dtrain2,
            self.dsim.copy(),
            periods=["2001", "2006"],
            xsdba_adjust_args={
                "detrend": {
                    "LoessDetrend": {"f": 0.2, "niter": 1, "d": 0, "weights": "tricube"}
                }
            },
        )

        assert (
            out.tas.attrs["bias_adjustment"]
            == "DetrendedQuantileMapping(group=Grouper(name='time.dayofyear',"
            " window=31), kind='+').adjust(sim, detrend=<LoessDetrend>),"
            " ref and hist were prepared with jitter_under_thresh(ref, hist,"
            " {'thresh': '2 K'}) and jitter_over_thresh(ref, hist, {'upper_bnd':"
            " '3 K', 'thresh': '2 K'}) and adapt_freq(ref, hist, {'thresh': '2 K'})"
        )

        assert (
            out2.tas.attrs["bias_adjustment"]
            == "DetrendedQuantileMapping(group=Grouper(name='time.dayofyear',"
            " window=31), kind='+').adjust(sim, detrend=<LoessDetrend>), ref and"
            " hist were prepared with jitter_under_thresh(ref, hist, {'thresh':"
            " '2 K'}) and jitter_over_thresh(ref, hist, {'upper_bnd': '3 K',"
            " 'thresh': '2 K'}) and adapt_freq(ref, hist, {'thresh': '2 K'})"
        )

        assert out.equals(out2)

    def test_xsdba_vs_xscen(
        self,
    ):  # should give the same results  using xscen and xclim
        dref = (
            timeseries(
                np.random.randint(0, high=10, size=(365 * 3) + 1),
                variable="pr",
                start="2001-01-01",
                freq="D",
                as_dataset=True,
            )
            .astype("float32")
            .chunk({"time": -1})
        )

        dsim = (
            timeseries(
                np.random.randint(0, high=10, size=365 * 6 + 1),
                variable="pr",
                start="2001-01-01",
                freq="D",
                as_dataset=True,
            )
            .astype("float32")
            .chunk({"time": -1})
        )
        dhist = dsim.sel(time=slice("2001", "2003")).chunk({"time": -1})

        # xscen version
        dtrain_xscen = xs.train(
            dref,
            dhist,
            var="pr",
            period=["2001", "2003"],
            adapt_freq={"thresh": "1 mm d-1"},
            xsdba_train_args={"kind": "*", "nquantiles": 50},
        )

        out_xscen = xs.adjust(
            dtrain_xscen,
            dsim,
            periods=["2001", "2006"],
            xsdba_adjust_args={
                "detrend": {
                    "LoessDetrend": {"f": 0.2, "niter": 1, "d": 0, "weights": "tricube"}
                },
                "interp": "nearest",
                "extrapolation": "constant",
            },
        )

        # xsdba version
        with xsdba.set_options(extra_output=True):
            group = xsdba.Grouper(group="time.dayofyear", window=31)

            drefx = dref.sel(time=slice("2001", "2003")).convert_calendar("noleap")
            dhistx = dhist.sel(time=slice("2001", "2003")).convert_calendar("noleap")
            dsimx = dsim.sel(time=slice("2001", "2006")).convert_calendar("noleap")

            # xsdba is now climate agnostics, hydro context must be specified
            with xc.core.units.units.context("hydro"):
                dhist_ad, pth, dP0 = xsdba.processing.adapt_freq(
                    drefx["pr"], dhistx["pr"], group=group, thresh="1 mm d-1"
                )

                QM = xsdba.DetrendedQuantileMapping.train(
                    drefx["pr"], dhist_ad, group=group, kind="*", nquantiles=50
                )

                detrend = xsdba.detrending.LoessDetrend(
                    f=0.2, niter=1, d=0, weights="tricube", group=group, kind="*"
                )
                out_xclim = QM.adjust(
                    dsimx["pr"],
                    detrend=detrend,
                    interp="nearest",
                    extrapolation="constant",
                ).rename({"scen": "pr"})

        assert out_xscen.equals(out_xclim)



import dask.array
import geopandas as gpd
import numpy as np
import pytest
import xarray as xr
import xclim as xc
from shapely.geometry import Polygon

import xscen as xs
from xscen.spatial import _estimate_grid_resolution, _load_lon_lat
from xscen.testing import datablock_3d


class TestCreepFill:
    # Create a 3D datablock
    ds = datablock_3d(
        np.tile(np.arange(1, 37).reshape(6, 6), (3, 1, 1)),
        "tas",
        "lon",
        -70,
        "lat",
        45,
        1,
        1,
        "2000-01-01",
        as_dataset=True,
    )
    ds["mask"] = ds["tas"].isel(time=0) > 0
    # Place a few False values in the mask
    ds["mask"][0, 0] = False
    ds["mask"][3, 3] = False

    @pytest.mark.parametrize(
        ("n", "mode"),
        [(1, "clip"), (2, "clip"), (3, "clip"), (1, "wrap"), (2, "wrap"), (3, "wrap")],
    )
    def test_n(self, n, mode):
        w = xs.spatial.creep_weights(self.ds["mask"], n=n, mode=mode)
        out = xs.spatial.creep_fill(self.ds["tas"], w)

        if mode == "clip":
            neighbours_0 = {
                1: [2, 7, 8],
                2: [2, 7, 8, 3, 9, 13, 14, 15],
                3: [
                    2,
                    7,
                    8,
                    3,
                    9,
                    13,
                    14,
                    15,
                    4,
                    10,
                    16,
                    19,
                    20,
                    21,
                ],  # 22 is False, thus not included
            }
            neighbours_3 = {
                # For these n, the average is the same as the original value
                1: [22],
                2: [22],
                # Here all the values are included, except the False ones
                3: [
                    (self.ds["tas"].isel(time=0).sum().values - 22 - 1)
                    / (self.ds["mask"].count().values - 2)
                ],
            }
        else:
            neighbours_0 = {
                1: [36, 31, 32, 6, 2, 12, 7, 8],
                2: [
                    29,
                    30,
                    25,
                    26,
                    27,
                    35,
                    36,
                    31,
                    32,
                    33,
                    5,
                    6,
                    2,
                    3,
                    11,
                    12,
                    7,
                    8,
                    9,
                    17,
                    18,
                    13,
                    14,
                    15,
                ],
                3: [
                    (
                        np.sum(np.arange(1, 37))
                        - 22
                        - 1
                        + np.sum([19, 20, 21, 23, 24])
                        + np.sum([4, 10, 16, 28, 34])
                    )
                    / (36 - 2 + 10)
                ],
            }
            neighbours_3 = {
                # For these n, the average is the same as the original value
                1: [22],
                2: [22],
                # Here all the values are included, except the False ones
                3: [
                    (
                        np.sum(np.arange(1, 37))
                        - 22
                        - 1
                        + np.sum(np.arange(2, 7))
                        + np.sum([7, 13, 19, 25, 31])
                    )
                    / (36 - 2 + 10)
                ],
            }

        np.testing.assert_allclose(
            out.isel(lat=0, lon=0), np.tile(np.mean(neighbours_0[n]), 3)
        )
        np.testing.assert_allclose(
            out.isel(lat=3, lon=3), np.tile(np.mean(neighbours_3[n]), 3)
        )

    def test_wrong_mode(self):
        with pytest.raises(ValueError, match="mode must be either"):
            xs.spatial.creep_weights(self.ds["mask"], n=1, mode="wrong")

    def test_n0(self):
        w = xs.spatial.creep_weights(self.ds["mask"], n=0, mode="clip")
        out = xs.spatial.creep_fill(self.ds["tas"], w)
        np.testing.assert_equal(out.isel(lat=0, lon=0), np.tile(np.nan, 3))
        np.testing.assert_equal(out.isel(lat=3, lon=3), np.tile(np.nan, 3))


class TestGetGrid:
    def test_none(self):
        ds = datablock_3d(
            np.zeros((20, 10, 10)),
            "tas",
            "lon",
            -142,
            "lat",
            0,
            2,
            2,
            "2000-01-01",
            as_dataset=True,
        )
        assert xs.spatial.get_grid_mapping(ds) == ""

    def test_rotated_pole(self):
        ds = datablock_3d(
            np.zeros((20, 10, 10)),
            "tas",
            "rlon",
            -142,
            "rlat",
            0,
            2,
            2,
            "2000-01-01",
            as_dataset=True,
        )
        assert xs.spatial.get_grid_mapping(ds) == "rotated_pole"

        ds_no_coord = ds.copy()
        ds_no_coord = ds_no_coord.drop_vars("rotated_pole")
        assert xs.spatial.get_grid_mapping(ds_no_coord) == "rotated_pole"

        ds_no_var = ds.copy()
        ds_no_var = ds_no_var.drop_vars("tas")
        assert xs.spatial.get_grid_mapping(ds_no_var) == "rotated_pole"

    def test_error(self):
        ds = datablock_3d(
            np.zeros((20, 10, 10)),
            "tas",
            "x",
            -5000,
            "y",
            5000,
            100000,
            100000,
            "2000-01-01",
            as_dataset=True,
        )
        ds["tas"].attrs["grid_mapping"] = "lambert_conformal_conic"
        with pytest.warns(
            UserWarning, match="There are conflicting grid_mapping attributes"
        ):
            assert xs.spatial.get_grid_mapping(ds) == "lambert_conformal_conic"


class TestSubset:
    ds = datablock_3d(
        np.ones((3, 50, 50)),
        "tas",
        "lon",
        -70,
        "lat",
        45,
        1,
        1,
        "2000-01-01",
        as_dataset=True,
    )

    @pytest.mark.parametrize(
        ("kwargs", "name"),
        [
            ({"lon": -70, "lat": 45}, None),
            ({"lon": [-53.3, -69.6], "lat": [49.3, 46.6]}, "foo"),
        ],
    )
    def test_subset_gridpoint(self, kwargs, name):
        with pytest.warns(UserWarning, match="tile_buffer is not used"):
            out = xs.spatial.subset(
                self.ds, "gridpoint", name=name, tile_buffer=5, **kwargs
            )

        if isinstance(kwargs["lon"], list):
            expected = {
                "lon": [np.round(k) for k in kwargs["lon"]],
                "lat": [np.round(k) for k in kwargs["lat"]],
            }
        else:
            expected = {
                "lon": [np.round(kwargs["lon"])],
                "lat": [np.round(kwargs["lat"])],
            }

        assert (
            f"gridpoint spatial subsetting on {len(expected['lon'])} coordinates"
            in out.attrs["history"]
        )
        np.testing.assert_array_equal(out["lon"], expected["lon"])
        np.testing.assert_array_equal(out["lat"], expected["lat"])
        if name:
            assert out.attrs["cat:domain"] == name
        else:
            assert "cat:domain" not in out.attrs

    @pytest.mark.parametrize(
        ("kwargs", "tile_buffer", "method"),
        [
            ({"lon_bnds": [-63, -60], "lat_bnds": [47, 50]}, 0, "bbox"),
            ({"lon_bnds": [-63, -60], "lat_bnds": [47, 50]}, 5, "bbox"),
            ({}, 0, "shape"),
            ({}, 5, "shape"),
            ({"buffer": 3}, 5, "shape"),
        ],
    )
    def test_subset_bboxshape(self, kwargs, tile_buffer, method):
        if method == "shape":
            gdf = gpd.GeoDataFrame(
                {"geometry": [Polygon([(-63, 47), (-63, 50), (-60, 50), (-60, 47)])]}
            )
            kwargs["shape"] = gdf

        if "buffer" in kwargs:
            with pytest.raises(
                ValueError, match="Both tile_buffer and clisops' buffer were requested."
            ):
                xs.spatial.subset(self.ds, method, tile_buffer=tile_buffer, **kwargs)
        else:
            out = xs.spatial.subset(self.ds, method, tile_buffer=tile_buffer, **kwargs)

            assert f"{method} spatial subsetting" in out.attrs["history"]
            if tile_buffer:
                assert f"with buffer={tile_buffer}" in out.attrs["history"]
                np.testing.assert_array_equal(
                    out["lon"],
                    np.arange(
                        np.max([-63 - tile_buffer, self.ds.lon.min()]),
                        np.min([-60 + tile_buffer + 1, self.ds.lon.max()]),
                    ),
                )
                np.testing.assert_array_equal(
                    out["lat"],
                    np.arange(
                        np.max([47 - tile_buffer, self.ds.lat.min()]),
                        np.min([50 + tile_buffer + 1, self.ds.lat.max()]),
                    ),
                )

            else:
                assert "with no buffer" in out.attrs["history"]
                np.testing.assert_array_equal(out["lon"], np.arange(-63, -59))
                np.testing.assert_array_equal(out["lat"], np.arange(47, 51))

    @pytest.mark.parametrize("crs", ["bad", "EPSG:3857", "EPSG:4326"])
    def test_shape_crs(self, crs):
        gdf = gpd.GeoDataFrame(
            {"geometry": [Polygon([(-63, 47), (-63, 50), (-60, 50), (-60, 47)])]}
        )
        if crs != "bad":
            gdf.crs = crs
            if crs != "EPSG:4326":
                with pytest.warns(UserWarning, match="Reprojecting to this CRS"):
                    with pytest.raises(
                        ValueError, match="No grid cell centroids"
                    ):  # This is from clisops, this is not our warning
                        xs.spatial.subset(self.ds, "shape", shape=gdf, tile_buffer=5)
            else:
                # Make sure there is no warning about reprojection
                with pytest.warns() as record:
                    xs.spatial.subset(self.ds, "shape", shape=gdf, tile_buffer=5)
                assert not any("Reprojecting to this CRS" in str(w) for w in record)

        else:
            with pytest.warns(UserWarning, match="does not have a CRS"):
                xs.spatial.subset(self.ds, "shape", shape=gdf, tile_buffer=5)

    def test_subset_sel(self):
        ds = datablock_3d(
            np.ones((3, 50, 50)),
            "tas",
            "rlon",
            -10,
            "rlat",
            0,
            1,
            1,
            "2000-01-01",
            as_dataset=True,
        )

        with pytest.raises(KeyError):
            xs.spatial.subset(ds, "sel", lon=[-75, -70], lat=[-5, 0])
        out = xs.spatial.subset(ds, "sel", rlon=[-5, 5], rlat=[0, 3])

        assert "sel subsetting" in out.attrs["history"]
        np.testing.assert_array_equal(out["rlon"], np.arange(-5, 6))
        np.testing.assert_array_equal(out["rlat"], np.arange(0, 4))

    def test_history(self):
        ds = self.ds.copy(deep=True)
        ds.attrs["history"] = "this is previous history"
        out = xs.spatial.subset(ds, "gridpoint", lon=-70, lat=45)

        assert "this is previous history" in out.attrs["history"].split("\n")[1]
        assert "gridpoint spatial subsetting" in out.attrs["history"].split("\n")[0]

    def test_subset_wrong_method(self):
        with pytest.raises(ValueError, match="Subsetting type not recognized"):
            xs.spatial.subset(self.ds, "wrong", lon=-70, lat=45)

    def test_subset_no_attributes(self):
        ds = self.ds.copy()
        ds.lat.attrs = {}
        ds.lon.attrs = {}
        assert "latitude" not in ds.cf

        xs.spatial.subset(
            ds,
            "bbox",
            name="test",
            lon_bnds=[-63, -60],
            lat_bnds=[47, 50],
        )


def test_dask_coords():
    ds = datablock_3d(
        np.ones((3, 50, 50)),
        "tas",
        "rlon",
        -10,
        "rlat",
        0,
        1,
        1,
        "2000-01-01",
        as_dataset=True,
    )
    # Transform the coordinates to dask arrays
    lon_attrs = ds["lon"].attrs
    ds["lon"] = xr.DataArray(
        dask.array.from_array(ds["lon"].data, chunks=(1, 1)),
        dims=ds["lon"].dims,
        attrs=lon_attrs,
    )
    lat_attrs = ds["lat"].attrs
    ds["lat"] = xr.DataArray(
        dask.array.from_array(ds["lat"].data, chunks=(1, 1)),
        dims=ds["lat"].dims,
        attrs=lat_attrs,
    )
    assert xc.core.utils.uses_dask(ds.cf["longitude"])

    ds = _load_lon_lat(ds)
    assert not xc.core.utils.uses_dask(ds.cf["longitude"])
    assert not xc.core.utils.uses_dask(ds.cf["latitude"])


@pytest.mark.parametrize(("lon_res", "lat_res"), [(0.5, 1), (1, 0.5)])
def test_estimate_res_1d(lon_res, lat_res):
    ds = datablock_3d(
        np.ones((3, 5, 5)),
        "tas",
        "lon",
        -10,
        "lat",
        0,
        lon_res,
        lat_res,
        "2000-01-01",
        as_dataset=True,
    )
    lon_res_est, lat_res_est = _estimate_grid_resolution(ds)
    assert lon_res_est == lon_res
    assert lat_res_est == lat_res


@pytest.mark.parametrize(("lon_res", "lat_res"), [(0.5, 1), (1, 0.5)])
def test_estimate_res_2d(lon_res, lat_res):
    ds = datablock_3d(
        np.ones((3, 5, 5)),
        "tas",
        "rlon",
        -10,
        "rlat",
        0,
        lon_res,
        lat_res,
        "2000-01-01",
        as_dataset=True,
    )
    lon_res_est, lat_res_est = _estimate_grid_resolution(ds)
    np.testing.assert_allclose(lon_res_est, ds.lon.diff("rlon").max())
    np.testing.assert_allclose(lat_res_est, ds.lat.diff("rlat").max())



import numpy as np
from conftest import notebooks
from xclim.testing.helpers import test_timeseries as timeseries

import xscen as xs
from xscen import scripting as sc


class TestScripting:
    ds = timeseries(
        np.tile(np.arange(1, 2), 50),
        variable="tas",
        start="2000-01-01",
        freq="YS-JAN",
        as_dataset=True,
    )
    ds.attrs = {
        "cat:type": "simulation",
        "cat:processing_level": "raw",
        "cat:source": "CanESM5",
        "cat:experiment": "ssp585",
        "cat:member": "r1i1p1f1",
        "cat:domain": "global",
        "cat:mip_era": "CMIP6",
        "cat:institution": "CCCma",
        "cat:activity": "ScenarioMIP",
    }

    def test_save_and_update(self):
        root = str(notebooks / "_data")

        cat = xs.ProjectCatalog.create(
            f"{root}/test_cat.json",
            project={"title": "tmp_cat", "id": "tmp"},
            overwrite=True,
        )

        first_path = root + "/test_{member}.zarr"
        sc.save_and_update(TestScripting.ds, cat, path=first_path)

        assert cat.df.path[0] == root + "/test_r1i1p1f1.zarr"
        assert cat.df.experiment[0] == "ssp585"

        sc.save_and_update(
            TestScripting.ds,
            cat,
            file_format="nc",
            # To fix hdf5 issues with h5py 3.11 on pip
            save_kwargs=dict(netcdf_kwargs={"engine": "netcdf4"}),
            build_path_kwargs={"root": root},
        )

        assert (
            cat.df.path[1]
            == root
            + "/simulation/raw/CMIP6/ScenarioMIP/global/CCCma/CanESM5/ssp585/r1i1p1f1/yr/tas/tas_yr_CMIP6_ScenarioMIP_global_CCCma_CanESM5_ssp585_r1i1p1f1_2000-2049.nc"  # noqa: E501
        )
        assert cat.df.source[1] == "CanESM5"

    def test_move_and_delete(self):
        root = str(notebooks / "_data")

        cat = xs.ProjectCatalog.create(
            f"{root}/test_cat.json",
            project={"title": "tmp_cat", "id": "tmp"},
            overwrite=True,
        )

        sc.save_and_update(
            TestScripting.ds, cat, path=root + "/dir1/test_r1i1p1f1.zarr"
        )
        sc.save_and_update(
            TestScripting.ds, cat, path=root + "/dir1/test_r1i1p1f2.zarr"
        )

        sc.move_and_delete(
            moving=[
                [root + "/dir1/test_r1i1p1f2.zarr", root + "/dir2/test_r1i1p1f2.zarr"]
            ],
            pcat=cat,
            deleting=[root + "/dir1"],
        )
        cat.update()

        # f2 should be moved to dir2 and dir1 should be deleted (not in row 0  anymore)
        assert cat.df.path[0] == root + "/dir2/test_r1i1p1f2.zarr"
        assert len(cat.df) == 1  # only one file left

        sc.move_and_delete(
            moving=[
                [root + "/dir2/test_r1i1p1f2.zarr", root + "/dir1/test_r1i1p1f2.zarr"]
            ],
            pcat=cat,
            copy=True,
        )
        cat.update()

        # f2 should be copied to dir1 and f1 should still exist
        assert cat.df.path[0] == root + "/dir2/test_r1i1p1f2.zarr"
        assert cat.df.path[1] == root + "/dir1/test_r1i1p1f2.zarr"
        assert len(cat.df) == 2  # only one file left



import numpy as np
from xclim.testing.helpers import test_timeseries as timeseries

import xscen.xclim_modules.conversions as conv


def test_precipitation():
    prsn = timeseries(np.ones(3), variable="prsn", start="2001-01-01", freq="D")
    prlp = timeseries(
        np.ones(3) * 10, variable="pr", start="2001-01-01", freq="D", units="mm d-1"
    )
    prlp.attrs["standard_name"] = "liquid_precipitation_flux"

    pr = conv.precipitation(prsn, prlp)
    assert pr.attrs["units"] == prsn.attrs["units"]
    np.testing.assert_array_equal(pr, prsn + prlp / 86400)


def test_tasmin_from_dtr():
    dtr = timeseries(
        np.ones(3) * 10, variable="tas", start="2001-01-01", freq="D", units="C"
    )
    dtr.attrs["standard_name"] = "daily_temperature_range"
    tasmax = timeseries(
        np.ones(3) * 20, variable="tasmax", start="2001-01-01", freq="D"
    )

    tasmin = conv.tasmin_from_dtr(dtr, tasmax)
    assert tasmin.attrs["units"] == tasmax.attrs["units"]
    np.testing.assert_array_equal(
        tasmin, tasmax - dtr
    )  # DTR in Celsius should not matter in the result, for a tasmax in Kelvin


def test_tasmax_from_dtr():
    dtr = timeseries(
        np.ones(3) * 10, variable="tas", start="2001-01-01", freq="D", units="C"
    )
    dtr.attrs["standard_name"] = "daily_temperature_range"
    tasmin = timeseries(
        np.ones(3) * 10, variable="tasmin", start="2001-01-01", freq="D"
    )

    tasmax = conv.tasmax_from_dtr(dtr, tasmin)
    assert tasmax.attrs["units"] == tasmin.attrs["units"]
    np.testing.assert_array_equal(
        tasmax, tasmin + dtr
    )  # DTR in Celsius should not matter in the result, for a tasmin in Kelvin


def test_dtr():
    tasmax = timeseries(
        np.ones(3) * 20, variable="tasmax", start="2001-01-01", freq="D", units="C"
    )
    tasmin = timeseries(
        np.ones(3) * 10, variable="tasmin", start="2001-01-01", freq="D"
    )

    dtr = conv.dtr_from_minmax(tasmin, tasmax)
    assert dtr.attrs["units"] == "°C"
    assert dtr.attrs["units_metadata"] == "temperature: difference"
    np.testing.assert_array_equal(dtr, (tasmax + 273.15) - tasmin)
